import argparse
import os
import re
from pathlib import Path


class DynamicCompose:
    CWD = None
    SAFELY_COVER_MARK = ('# This file was automatically generated by DynamicCompose, and can be safely deleted, '
                         'but do not edit it or remove this comment.\n')

    def __init__(self, show_command: bool = True, force_cover: bool = False):
        self.show_command = show_command
        self.force_cover = force_cover
        self.cwd = self.CWD or Path().cwd()
        self.env_file = self.cwd / '.env'
        self.env_file.touch(exist_ok=True)
        self.compose_dir = self.cwd / '.compose'
        self.compose_dir.mkdir(exist_ok=True)
        self.compose_file = self.cwd / 'compose.yml'
        self.dockerfile_file = self.cwd / 'Dockerfile'
        self.compose_content = ''
        self.dockerfile_content = ''

        self.env = self._get_env()
        args, self.extra_args = self._get_args()
        self.env.update(dict(args.e or ()))
        self.arg = {}  # Args used in building

    @staticmethod
    def _parse_key_value(arg_string: str):
        key, value = arg_string.strip().split('=', maxsplit=1)
        return key, value

    def _get_env(self):
        env_dict = {}
        if self.env_file.exists():
            with open(self.env_file) as f:
                for line in f.readlines():
                    if line.strip():
                        key, value = self._parse_key_value(line)
                        env_dict[key] = value
        return env_dict

    def _get_args(self) -> tuple[argparse.Namespace, list[str]]:
        parser = argparse.ArgumentParser(description='Dynamic generating the Dockerfile and the compose.yml')
        parser.add_argument('-e', action='append', type=self._parse_key_value,
                            help='The environment variables are used in generating.')
        parsed_args, parsed_extra_args = parser.parse_known_args()
        return parsed_args, parsed_extra_args

    @staticmethod
    def _convert_env_var_ref(template: str) -> str:
        return re.sub(r'\$\{(.+?)}', r'${{\1}}', template)

    def format_compose(self, template: str = None, template_file: str | Path = None) -> str:
        if not template:
            template = Path(template_file or self.compose_dir / 'compose.yml').read_text(encoding='utf-8')
        self.compose_content = self._convert_env_var_ref(template).format(**self.env, **self.arg)
        return self.compose_content

    def format_dockerfile(self, template: str = None, template_file: str | Path = None) -> str:
        if not template:
            template = Path(template_file or self.compose_dir / 'Dockerfile').read_text(encoding='utf-8')
        self.dockerfile_content = self._convert_env_var_ref(template).format(**self.env, **self.arg)
        return self.dockerfile_content

    def dump_compose(self, target_file: str | Path = None) -> None:
        Path(target_file or self.compose_file).write_text(self.SAFELY_COVER_MARK + self.compose_content,
                                                          encoding='utf-8')

    def dump_dockerfile(self, target_file: str | Path = None) -> None:
        Path(target_file or self.dockerfile_file).write_text(self.SAFELY_COVER_MARK + self.dockerfile_content,
                                                             encoding='utf-8')

    def fd_compose(self, template: str = None, template_file: str | Path = None,
                   target_file: str | Path = None) -> None:
        self.format_compose(template, template_file)
        self.dump_compose(target_file)

    def fd_dockerfile(self, template: str = None, template_file: str | Path = None,
                      target_file: str | Path = None) -> None:
        self.format_dockerfile(template, template_file)
        self.dump_dockerfile(target_file)

    def dump(self) -> None:
        self.dump_compose()
        self.dump_dockerfile()

    def up(self, with_extra_args: bool = True) -> int:
        command = f'docker compose up'
        if with_extra_args:
            command = f'{command} {" ".join(self.extra_args)}'
        if self.show_command:
            print(command)
        return os.system(command)



import time
from threading import Thread
from tests.__init__ import *
from psplpy.other_utils import *


def test_is_sys():
    assert is_sys(is_sys.LINUX) is True
    assert is_sys(is_sys.WINDOWS) is False


def test_recursive_convert():
    data = [1, (2, [3, 4])]
    assert recursive_convert(data, to=list) == [1, [2, [3, 4]]]
    assert recursive_convert(data, to=tuple) == (1, (2, (3, 4)))


def test_split_list():
    lst = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    assert split_list(lst, 4) == [[1, 2, 3], [4, 5], [6, 7], [8, 9]]


def test_get_key_from_value():
    dct = {'a': 1, 'b': 2, 'c': 1, 'd': 1}
    assert get_key_from_value(dct, 1) == 'a'
    assert get_key_from_value(dct, 2, find_all=True) == ['b']
    assert get_key_from_value(dct, 1, find_all=True) == ['a', 'c', 'd']
    assert get_key_from_value(dct, 3, allow_null=True) is None
    assert get_key_from_value(dct, 3, find_all=True, allow_null=True) is None
    try:
        get_key_from_value(dct, 3)
    except KeyError as e:
        print(e)
    else:
        assert False


def test_get_env():
    assert get_env('SERVICE') == 'psplpy'
    try:
        get_env('S')
    except KeyError as e:
        pass
    else:
        raise AssertionError


def test_perf_counter():
    p = PerfCounter()
    time.sleep(0.1)
    assert int(p.elapsed()) == int(0.1 * 1000)
    time.sleep(0.2)
    p.show('perf')


def test_timeout_checker():
    def checker(timeout: float, unique_id: Any = None, check_time: float = None,
                func: str = 'ret_false', end_loop_time: float = 10000):
        t_start = time.time()
        while getattr(TimeoutChecker(timeout, unique_id=unique_id), func)():
            if time.time() - t_start > end_loop_time:
                break
            time.sleep(0.01)
        print(time.time() - t_start)
        assert int(time.time() - t_start) == check_time or timeout

    Thread(target=checker, kwargs={'timeout': 1, 'unique_id': 1}).start()
    timeout = 2
    Thread(target=checker, kwargs={'timeout': 2, 'unique_id': 2}).start()
    time.sleep(timeout + 0.5)

    kwargs = {'timeout': 2, 'unique_id': 1, 'check_time': 1, 'func': 'raise_err', 'end_loop_time': 1}
    Thread(target=checker, kwargs=kwargs).start()
    try:
        checker(timeout=1, unique_id=2, func='raise_err')
    except Exception as e:
        assert isinstance(e, TimeoutError)


def test_screenshot_maker():
    pass
    # ScreenShotMaker()


def tests():
    print('test other utils')
    test_is_sys()
    test_recursive_convert()
    test_split_list()
    test_get_key_from_value()
    test_get_env()
    test_timeout_checker()
    test_screenshot_maker()
    test_perf_counter()


if __name__ == '__main__':
    tests()



from tests.__init__ import *
from psplpy.image_det import *
from psplpy.ocr_utils import *
from tests.test_ocr_utils import run_ocr_server, host


def test_img_det():
    image = rc_dir / 'cat.png'
    template_image = rc_dir / 'det_block.png'

    p = PerfCounter()
    result = ImgDet.get_boxes(image, template_image, confidence=0.8, method=ImgDet.TM_CCOEFF_NORMED,
                              draw_result=tmp_dir / 'det_marked.png', save_match=tmp_dir / 'det_match.png')
    p.show('det')
    print(result)
    result = ImgDet.get_box(image, template_image, confidence=0.8, method=ImgDet.TM_CCOEFF_NORMED,
                            draw_result=tmp_dir / 'det_marked.png', save_match=tmp_dir / 'det_match.png')
    p.show('det')
    print(result)
    assert result == (Rect([(840, 189), (890, 239)], rect_format=Rect.LT_RB), 0.9999983906745911)


def test_txt_det():
    run_ocr_server()
    image = rc_dir / 'ocr.png'
    p = PerfCounter()
    ocr_result = Ocr(host=host).get(image)
    p.show('ocr')
    print(ocr_result)

    def do_noting(txt: str) -> str:
        return txt

    t = TxtDet(txt_processor=TxtProcessor(TxtProcessor.NO_SPACE, TxtProcessor.CASEFOLD, do_noting),
               process_text=True)
    boxes = t.get_boxes(ocr_result, '右上角Test')
    p.show('boxes')
    print(boxes)
    assert boxes == TextInfoList([[Rect([[773.0, 6.0], [992.0, 28.0]]), '我是右上角Test文本', 0.9980501532554626]])
    text = t.get_txt(ocr_result, Rect([[773.0, 6.0], [992.0, 28.0]]))
    p.show('text')
    print(text)
    match_obj = t.match(ocr_result, [[Rect([[773.0, 6.0], [992.0, 28.0]]), '右上角Test', 0.99],
                                     [Rect([[100, 6.0], [200, 28.0]]), 'mismatch文本', 0.99]], confidence=0.5)
    p.show('match_obj')
    assert (match_obj.bool, match_obj.result_list, match_obj.confidence) == (True, [True, False], 0.5)
    assert match_obj.match_list == [[Rect([[773.0, 6.0], [992.0, 28.0]]), '右上角Test', 0.99]]
    assert match_obj.mismatch_list == [[Rect([[100, 6.0], [200, 28.0]]), 'mismatch文本', 0.99]]


def tests():
    test_img_det()
    test_txt_det()


if __name__ == '__main__':
    tests()



import threading
import time
import pyautogui
import pyperclip
from pynput.keyboard import Controller, Key
from psplpy.concurrency_utils import ThreadLocks
from psplpy.network_utils import ServerSocket, ClientSocket


pyautogui.FAILSAFE = False
keyboard = Controller()


class AutoGui:
    def paste(self, s: str, pos: tuple[int, int] = (), desktop_pos: tuple[int, int] = (pyautogui.size()[0], 0)):
        for i in range(2):
            pyperclip.copy(s)
            pyautogui.click(*desktop_pos)
            time.sleep(0.1)
            pyautogui.click(*pos)
            time.sleep(0.1)
        keyboard.press(Key.ctrl)
        keyboard.press('v')
        time.sleep(0.1)
        pyautogui.click(1, 1)
        time.sleep(0.1)
        keyboard.release(Key.ctrl)
        keyboard.release('v')

    @staticmethod
    def drag_from(start_x: int, start_y: int, end_x: int, end_y: int, duration: float = 1):
        pyautogui.moveTo(start_x, start_y)
        return pyautogui.dragTo(end_x, end_y, duration=duration)


class AutoGuiCommand:
    def __init__(self, command: str, args: tuple | list, kwargs: dict, id: int, send_time: float):
        self.command = command
        self.args = args
        self.kwargs = kwargs
        self.id = id
        self.send_time = send_time  # the time when the command is sent
        self.exec_time = 0  # the execution time of the command
        self.result = None
        self.elapsed = 0  # the elapsed time between sending command and receiving result

    @property
    def tour_time(self) -> float:  # the time of network latency
        return self.elapsed - self.exec_time

    def __str__(self):
        return f"{str(self.__dict__)[:-1]}, 'tour_time': {self.tour_time}}}"

    __repr__ = __str__


class AutoGuiCommandNotFoundError(Exception):
    def __init__(self, command: str):
        self.command = command

    def __str__(self):
        return f"AutoGuiCommand not found: {self.command}"


class AutoGuiSender:
    def __init__(self, host: str = '127.0.0.1', port: int = 12345):
        self._client_socket = ClientSocket(host=host, port=port)
        self._client_socket.connect()
        self._id = 0

    def send(self, command: str, args: tuple | list = None, kwargs: dict = None) -> AutoGuiCommand:
        ac = AutoGuiCommand(command, args or (), kwargs or {}, self._id, time.time())
        self._id += 1
        self._client_socket.send_pickle(ac)
        ac: AutoGuiCommand = self._client_socket.recv_pickle()
        ac.elapsed = time.time() - ac.send_time
        if isinstance(ac.result, Exception):
            raise ac.result
        return ac


class AutoGuiExecutor:
    KM_LOCK = '_keyboard_and_mouse_lock'

    def __init__(self, host: str = '0.0.0.0', port: int = 12345):
        self._locks = ThreadLocks()
        self._locks.set_lock(self.KM_LOCK, auto_release_time=300)
        self._autogui = AutoGui()
        self._server_socket = ServerSocket(host=host, port=port)

    def run(self, new_thread: bool = False) -> None:
        if new_thread:
            threading.Thread(target=self._server_socket.handle, args=(self._handler,)).start()
        else:
            self._server_socket.handle(self._handler)

    def _handler(self, client_socket: ClientSocket, address):
        objs_require_km = [self._autogui, pyautogui]
        while True:
            ac: AutoGuiCommand = client_socket.recv_pickle()
            if not ac:
                break
            t_start = time.perf_counter()
            for obj in [self._locks, *objs_require_km]:
                if ac.command in dir(obj):
                    if obj in objs_require_km:
                        self._locks.acquire_lock(self.KM_LOCK)
                    result = getattr(obj, ac.command)(*ac.args, **ac.kwargs)
                    if obj in objs_require_km:
                        self._locks.release_lock(self.KM_LOCK)
                    break
            else:
                result = AutoGuiCommandNotFoundError(ac.command)
            ac.exec_time = time.perf_counter() - t_start
            ac.result = result
            client_socket.send_pickle(ac)



import math
from pathlib import Path
from typing import Any
from PIL import Image
from psplpy.image_utils import ImgConv, ImgDraw, Rect
from psplpy.network_utils import MpHttpServer, MpHttpClient


class OcrServer(MpHttpServer):
    def __init__(self, host: str = '0.0.0.0', port: int = 80, workers: int = 1, show_info: bool = True,
                 max_fetch_timeout: float = 3600, result_timeout: float = 3600,
                 use_gpu: bool = False, lang: str = 'ch'):
        self.use_gpu, self.lang = use_gpu, lang
        self._import_paddleocr(not self.use_gpu)
        super().__init__(host, port, workers, show_info, max_fetch_timeout, result_timeout)

    @staticmethod
    def _import_paddleocr(whether_to_import: bool = True):
        if whether_to_import:
            from paddleocr import PaddleOCR
            global PaddleOCR
            PaddleOCR = PaddleOCR

    @staticmethod
    def unify_format(ocr_result: list) -> list:
        """format text info list to: [[[[left top], [right top], [right bottom], [left bottom]], text, confidence], ...]
           for instance:             [[[[49, 5], [89, 5], [89, 19], [49, 19]], 'Text', 0.0004511636577615576], ...]"""
        if ocr_result:  # 防止传入列表本身为空
            ocr_result = ocr_result[0]
            if ocr_result is not None:  # paddleocr如果未检测到，会返回 [None]
                return [[text_info[0], text_info[1][0], text_info[1][1]] for text_info in ocr_result]
        return []

    def init(self) -> None:
        self._import_paddleocr(self.use_gpu)
        self.ocr = PaddleOCR(det_model_dir='/home/a/.paddleocr/whl/det/ch/ch_PP-OCRv4_det_infer',
                             rec_model_dir='/home/a/.paddleocr/whl/rec/ch/ch_PP-OCRv4_rec_infer',
                             cls_model_dir='/home/a/.paddleocr/whl/cls/ch_ppocr_mobile_v2.0_cls_infer',
                             use_gpu=self.use_gpu, use_angle_cls=True, lang=self.lang, show_log=self.show_info)

    def main_loop(self, data: dict) -> Any:
        result = self.ocr.ocr(ImgConv(data['img']).to_numpy(), cls=data.get('cls', False))
        return self.unify_format(result)


class Ocr(MpHttpClient):
    def submit(self, imgs: list | tuple, cls: bool = False) -> int:
        data_list = []
        not_local = self.host not in ['localhost', '127.0.0.1']
        for img in imgs:
            if not_local and isinstance(img, (str, Path)):
                img = ImgConv(img).to_bytesio()
            data_list.append({'img': img, 'cls': cls})
        return super().submit(data_list)

    def fetch(self, task_id: int, timeout: float = 3600) -> list[list]:
        return super().fetch(task_id, timeout)

    def batch(self, imgs: list | tuple, cls: bool = False, timeout: float = 3600) -> list[list]:
        return self.fetch(self.submit(imgs, cls), timeout)

    def get(self, img: Any, cls: bool = False, timeout: float = 3600) -> list:
        return self.batch([img], cls, timeout)[0]


class DrawResult:
    def __init__(self, img: Any, ocr_result: list, save_path: Path | str = None, font_size_ratio: float = 0.7):
        self.img = ImgConv(img).to_pil()
        self.ocr_result, self.save_path, self.font_size_ratio = ocr_result, save_path, font_size_ratio
        self._draw()

    def _draw(self) -> Image.Image | Path:
        if self.ocr_result:
            self.img = ImgDraw(self.img).polygons([i[0] for i in self.ocr_result], outline='red').to_pil()
            self.img = self._draw_text()
        if self.save_path:
            return ImgConv(self.img).to_path(self.save_path)
        return self.img

    def _draw_text(self) -> Image.Image:
        w, h = self.img.size
        draw = ImgDraw(self.img)
        for text_info in self.ocr_result:
            quad = text_info[0]
            edges = (math.sqrt((quad[0][0] - quad[1][0]) ** 2 + (quad[0][1] - quad[1][1]) ** 2),
                     math.sqrt((quad[1][0] - quad[2][0]) ** 2 + (quad[1][1] - quad[2][1]) ** 2),
                     math.sqrt((quad[2][0] - quad[3][0]) ** 2 + (quad[2][1] - quad[3][1]) ** 2),
                     math.sqrt((quad[3][0] - quad[0][0]) ** 2 + (quad[3][1] - quad[0][1]) ** 2))
            min_edge_length = min(edges)
            max_edge_length = max(edges)
            bounding_rect = ImgDraw.bounding_rect(quad).to_lt_rb()

            # make the text rect not overflow the image edge
            right_x = bounding_rect[0][0] + max_edge_length
            bottom_y = bounding_rect[1][1] + min_edge_length
            x_diff = right_x - w if right_x > w else 0
            y = bounding_rect[0][1] - min_edge_length if bottom_y > h else bounding_rect[1][1]
            text_rect = Rect([bounding_rect[0][0] - x_diff, y,
                              max_edge_length, min_edge_length], rect_format=Rect.LTWH)
            draw.text_in_rect(text_info[1], text_rect, 'red')
        return draw.to_pil()



import os
import re
from pathlib import Path
import pyshark
from serialization_utils import Serializer


class BlockJetBrains:
    def __init__(self, interface: str):
        self.interface = interface

    @staticmethod
    def _extract_substring(string: str, search_string: str):
        result = []
        start = 0
        while True:
            index = string.find(search_string, start)
            if index == -1:
                break
            # extract the search string until to the \n
            end_index = string.find('\n', index)
            if end_index == -1:
                end_index = len(string)
            substring = string[index + len(search_string):end_index].strip()
            result.append(substring)
            start = end_index + 1
        return result

    @staticmethod
    def _p_os_system(command: str) -> None:
        print(command)
        os.system(command)

    def _flush_dns(self):
        self._p_os_system('ipconfig /flushdns')

    @staticmethod
    def _remove_ansi_escapes(text):
        ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
        return ansi_escape.sub('', text)

    def append_dns_ip(self):
        self._flush_dns()
        s = Serializer(Path(__file__), data_type=dict, embedded=True)
        current_ip_dict = s.load_json()
        print(current_ip_dict)
        capture_filter = 'dns'
        capture = pyshark.LiveCapture(interface=self.interface, display_filter=capture_filter)
        capture.sniff(timeout=1)

        print('Start capture')
        for packet in capture.sniff_continuously():
            has_added = False
            content = self._remove_ansi_escapes(str(packet))
            if 'Standard query response' in content:
                name_list = self._extract_substring(content, 'Name: ')
                name = None
                if name_list:
                    name = name_list[0]
                ip_result = self._extract_substring(content, ' addr ')
                print(f'{name}: {ip_result}')
                if name in ['www.jetbrains.com', 'account.jetbrains.com']:
                    for ip in ip_result:
                        if not current_ip_dict.get(name):
                            current_ip_dict[name] = []
                        current_ip_list = current_ip_dict[name]
                        if ip not in current_ip_list:
                            current_ip_list.append(ip)
                            print(f'Appended: {ip}')
                            has_added = True
                    if has_added:
                        s.dump_json(current_ip_dict)

                        rule_name = 'jetbrains'
                        total_ip_list = []
                        for ip_list in current_ip_dict.values():
                            total_ip_list.extend(ip_list)
                        self._p_os_system(f'netsh advfirewall firewall delete rule name="{rule_name}"')
                        self._p_os_system(f'netsh advfirewall firewall add rule name="{rule_name}" '
                                          f'dir=out action=block remoteip={",".join(total_ip_list)}')
                        self._p_os_system(f'netsh advfirewall firewall add rule name="{rule_name}" '
                                          f'dir=in action=block remoteip={",".join(total_ip_list)}')
                        self._flush_dns()
        capture.close()


if __name__ == '__main__':
    interface_name = 'WLAN'
    BlockJetBrains(interface_name).append_dns_ip()

"""2qIgKf9iDsVrsTMS
{
    "www.jetbrains.com": [
        "13.225.183.89",
        "13.225.183.104",
        "13.225.183.35",
        "13.225.183.60",
        "65.8.161.4",
        "65.8.161.125",
        "65.8.161.122",
        "65.8.161.71",
        "108.138.85.120",
        "108.138.85.118",
        "108.138.85.7",
        "108.138.85.106",
        "216.137.39.88",
        "216.137.39.118",
        "216.137.39.11",
        "216.137.39.60",
        "18.66.102.125",
        "18.66.102.36",
        "18.66.102.54",
        "18.66.102.75",
        "108.156.133.13",
        "108.156.133.59",
        "108.156.133.65",
        "108.156.133.6"
    ],
    "account.jetbrains.com": [
        "76.223.63.197",
        "13.248.188.196",
        "108.128.182.4",
        "34.248.135.53"
    ]
}
2qIgKf9iDsVrsTMS"""


from tests.__init__ import *
from psplpy.file_utils import *


def test_auto_rename():
    cat_path = rc_dir / 'cat.png'
    renamed_cat_path = str(cat_path).replace('.png', '(1).png')
    print(renamed_cat_path)
    assert auto_rename(cat_path) == Path(renamed_cat_path)
    assert auto_rename(str(cat_path)) == renamed_cat_path


def test_get_file_paths():
    folder_path = rc_dir / 'test_get_paths'

    abs_list = ['/home/a/src/psplpyProject/tests/resources/test_get_paths/test.txt',
                '/home/a/src/psplpyProject/tests/resources/test_get_paths/inner/inner_test.txt',
                '/home/a/src/psplpyProject/tests/resources/test_get_paths/inner/inner_test2.txt']
    rel_list = ['test.txt', 'inner/inner_test.txt', 'inner/inner_test2.txt']

    paths_list = get_file_paths(folder_path)
    assert paths_list == abs_list

    paths_list = get_file_paths(folder_path, relative=True)
    assert paths_list == rel_list

    paths_list = get_file_paths(folder_path, generator=True)
    assert isinstance(paths_list, Generator), type(paths_list)
    assert list(paths_list) == abs_list


def tests():
    print('test file utils')
    test_auto_rename()
    test_get_file_paths()


if __name__ == '__main__':
    tests()



import io
from pathlib import Path
from typing import Any, Callable, Union
import cv2
import numpy as np
from PIL import Image, ImageDraw, ImageEnhance, ImageOps, ImageFont
from psplpy.other_utils import recursive_convert, is_sys


class Rect:
    RECT = 'rect'                   # the Rect object
    LTWH = 'ltwh'                   # (left, top, width, height)
    LT_RB = 'lt_rb'                 # ((left, top), (right, bottom))
    LTRB = 'ltrb'                   # (left, top, right, bottom)
    LT_RT_RB_LB = 'lt_rt_rb_lb'     # ((left, top), (right, top), (right, bottom), (left, bottom))
    CENTER = 'center'               # (x, y)
    CENTER_RECT = 'center_rect'     # (center_x, center_y, half_width, half_height)

    def __init__(self, rect: Union['Rect', tuple, list], rect_format: str = None):
        rect_format = rect_format or self._det_rect_format(rect)
        self.rect = self._to_lt_rb(rect, rect_format)

    @staticmethod
    def _det_rect_format(rect: Any) -> str:
        if isinstance(rect, Rect):
            return Rect.RECT
        else:
            if len(rect) == 4:
                if isinstance(rect[0], (int, float)):
                    return Rect.LTWH
                elif len(rect[0]) == 2:
                    return Rect.LT_RT_RB_LB
            elif len(rect) == 2:
                return Rect.LT_RB
            raise ValueError(f'Unknown rectangle type for {rect}.')

    def _to_lt_rb(self, rect: Any, rect_format: str) -> list[list[float]]:
        if rect_format == Rect.RECT:
            result = rect.to_lt_rb()
        elif rect_format == Rect.LT_RB:
            self.rect = rect
            result = self.to_lt_rb()
        elif rect_format == Rect.LTWH:
            left, top, width, height = rect
            result = [[left, top], [left + width, top + height]]
        elif rect_format == Rect.LTRB:
            left, top, right, bottom = rect
            result = [[left, top], [right, bottom]]
        elif rect_format == Rect.LT_RT_RB_LB:
            left, top, right, bottom = rect[0][0], rect[0][1], rect[2][0], rect[2][1]
            result = [[left, top], [right, bottom]]
        elif rect_format == Rect.CENTER_RECT:
            center_x, center_y, half_width, half_height = rect
            left = center_x - half_width
            top = center_y - half_height
            right = center_x + half_width
            bottom = center_y + half_height
            result = [[left, top], [right, bottom]]
        else:
            raise ValueError(f"Unsupported rectangle format {rect_format}")
        return result

    def __iter__(self):
        for item in self.rect:
            yield item

    def __getitem__(self, index) -> list[float]:
        return self.rect[index]

    def __setitem__(self, index, value: list[float] | tuple[float, float]) -> 'Rect':
        self.rect[index] = list(value)
        return self

    def __str__(self):
        return str(f'Rect{self.to_lt_rb()}')

    __repr__ = __str__

    def __eq__(self, other: 'Rect'):
        if isinstance(other, Rect):
            return self.rect == other.rect
        return False

    def _to_other(self, rect_format: str) -> list:
        (left, top), (right, bottom) = self.rect
        if rect_format == Rect.LTWH:
            width = right - left
            height = bottom - top
            return [left, top, width, height]
        elif rect_format == Rect.LTRB:
            return [left, top, right, bottom]
        elif rect_format == Rect.LT_RT_RB_LB:
            return [[left, top], [right, top], [right, bottom], [left, bottom]]
        elif rect_format == Rect.CENTER:
            center_x = (left + right) / 2
            center_y = (top + bottom) / 2
            return [center_x, center_y]
        elif rect_format == Rect.CENTER_RECT:
            center_x = (left + right) / 2
            center_y = (top + bottom) / 2
            half_width = (right - left) / 2
            half_height = (bottom - top) / 2
            return [center_x, center_y, half_width, half_height]
        else:
            raise ValueError(f"Unsupported rectangle format {rect_format}")

    def to_lt_rb(self) -> list[list[float]]:
        return [[self.rect[0][0], self.rect[0][1]], [self.rect[1][0], self.rect[1][1]]]

    def to_ltwh(self) -> list[float]:
        return self._to_other(Rect.LTWH)

    def to_ltrb(self) -> list[float]:
        return self._to_other(Rect.LTRB)

    def to_lt_rt_rb_lt(self) -> list[list[float]]:
        return self._to_other(Rect.LT_RT_RB_LB)

    def to_center(self) -> list[float]:
        return self._to_other(Rect.CENTER)

    def to_center_rect(self) -> list[float]:
        return self._to_other(Rect.CENTER_RECT)

    def resize(self, scale: float) -> 'Rect':
        (left, top), (right, bottom) = self.rect

        width = right - left
        height = bottom - top

        new_width = width * scale
        new_height = height * scale

        new_left = left + (width - new_width) / 2
        new_top = top + (height - new_height) / 2
        new_right = new_left + new_width
        new_bottom = new_top + new_height

        self.rect = [[new_left, new_top], [new_right, new_bottom]]
        return self

    def area(self) -> float:
        (left, top), (right, bottom) = self.rect
        width = right - left
        height = bottom - top
        area = width * height
        return area

    def is_inside(self, internal_rect: 'Rect') -> bool:
        """whether a rect is inside the self_rect"""
        self_rect, other_rect = self.to_lt_rb(), internal_rect.to_lt_rb()
        (ax1, ay1), (ax2, ay2) = self_rect
        (bx1, by1), (bx2, by2) = other_rect
        return bx1 >= ax1 and bx2 <= ax2 and by1 >= ay1 and by2 <= ay2

    def overlap_pct(self, rect: 'Rect') -> tuple[float, float]:
        """calculate the ratio of the overlapping of two rectangles to the area of each rectangle"""
        self_rect, other_rect = self.to_lt_rb(), rect.to_lt_rb()

        (x1, y1), (r1, b1) = self_rect
        (x2, y2), (r2, b2) = other_rect
        x_overlap = max(0, min(r1, r2) - max(x1, x2))
        y_overlap = max(0, min(b1, b2) - max(y1, y2))

        area_overlap = x_overlap * y_overlap
        percent1 = area_overlap / self.area()
        percent2 = area_overlap / rect.area()
        return percent1, percent2


class ImgConv:
    def __init__(self, image: Image.Image | np.ndarray | bytes | str | Path | io.BytesIO,
                 from_opencv: bool = False):
        self.image = image
        if from_opencv:
            self.image = cv2.cvtColor(self.image, cv2.COLOR_BGR2RGB)

    def _raise_error(self):
        raise ValueError(f"Unsupported input data type {type(self.image)}")

    def to_numpy(self) -> np.ndarray:
        if isinstance(self.image, Image.Image):
            return np.array(self.image)
        elif isinstance(self.image, np.ndarray):
            return self.image
        elif isinstance(self.image, (str, Path, io.BytesIO)):
            return np.array(self.to_pil())
        elif isinstance(self.image, (bytes, bytearray)):
            return ImgConv(self.to_bytesio()).to_numpy()
        else:
            self._raise_error()

    def to_pil(self) -> Image.Image:
        if isinstance(self.image, Image.Image):
            return self.image
        elif isinstance(self.image, np.ndarray):
            return Image.fromarray(self.image)
        elif isinstance(self.image, (str, Path, io.BytesIO)):
            return Image.open(self.image)
        elif isinstance(self.image, (bytes, bytearray)):
            return ImgConv(self.to_bytesio()).to_pil()
        else:
            self._raise_error()

    def to_opencv(self) -> np.ndarray:
        if isinstance(self.image, Image.Image):
            return cv2.cvtColor(np.array(self.image), cv2.COLOR_RGB2BGR)
        elif isinstance(self.image, np.ndarray):
            return cv2.cvtColor(self.image, cv2.COLOR_RGB2BGR)
        elif isinstance(self.image, (str, Path)):
            return cv2.imread(str(self.image))
        elif isinstance(self.image, io.BytesIO):
            return cv2.cvtColor(self.to_numpy(), cv2.COLOR_RGB2BGR)
        elif isinstance(self.image, (bytes, bytearray)):
            return ImgConv(self.to_bytesio()).to_opencv()
        else:
            self._raise_error()

    @staticmethod
    def _get_save_params(**kwargs) -> dict:
        return kwargs

    def to_path(self, path: str | Path = None) -> Path:
        if isinstance(self.image, Image.Image):
            self.image.save(**self._get_save_params(fp=path))
            return Path(path)
        elif isinstance(self.image, np.ndarray):
            Image.fromarray(self.image).save(**self._get_save_params(fp=path))
            return Path(path)
        elif isinstance(self.image, (str, Path)):
            if not path:
                return Path(self.image)
            return ImgConv(self.to_pil()).to_path(path)
        elif isinstance(self.image, io.BytesIO):
            Image.open(self.image).save(**self._get_save_params(fp=path))
            return Path(path)
        elif isinstance(self.image, (bytes, bytearray)):
            return ImgConv(self.to_bytesio()).to_path(path)
        else:
            self._raise_error()

    def to_bytesio(self) -> io.BytesIO:
        if isinstance(self.image, Image.Image):
            bytes_io = io.BytesIO()
            self.image.save(bytes_io, **self._get_save_params(format='PNG'))
            return bytes_io
        elif isinstance(self.image, np.ndarray):
            bytes_io = io.BytesIO()
            Image.fromarray(self.image).save(bytes_io, **self._get_save_params(format='PNG'))
            return bytes_io
        elif isinstance(self.image, (str, Path)):
            bytes_io = io.BytesIO()
            Image.open(self.image).save(bytes_io, **self._get_save_params(format='PNG'))
            return bytes_io
        elif isinstance(self.image, io.BytesIO):
            return self.image
        elif isinstance(self.image, (bytes, bytearray)):
            return io.BytesIO(self.image)
        else:
            self._raise_error()

    def to_bytes(self) -> bytes:
        return self.to_bytesio().getvalue()

    def to_bytearray(self) -> bytearray:
        return bytearray(self.to_bytes())


class Img(ImgConv):
    def __init__(self, image: Image.Image | np.ndarray | bytes | str | Path | io.BytesIO,
                 from_opencv: bool = False):
        super().__init__(image, from_opencv)
        if isinstance(image, (str, Path)):
            self.fp = image
        self.image = self.to_pil()

    def resize(self, ratio: float = None, x: float = None, y: float = None) -> 'Img':
        if ratio:
            w_h = (int(self.image.width * ratio), int(self.image.height * ratio))
            self.image = self.image.resize(w_h, Image.BICUBIC)
        elif x and y:
            self.image = self.image.resize((int(x), int(y)), Image.BICUBIC)
        return self

    def _enhance(self, enhance_type, *args) -> 'Img':
        enhancer = getattr(ImageEnhance, enhance_type)(self.image)
        self.image = enhancer.enhance(*args)
        return self

    def contrast(self, contrast: float) -> 'Img':
        return self._enhance("Contrast", contrast)

    def brightness(self, brightness: float) -> 'Img':
        return self._enhance("Brightness", brightness)

    def sharpness(self, sharpness: float) -> 'Img':
        return self._enhance("Sharpness", sharpness)

    def grayscale(self) -> 'Img':
        self.image = self.image.convert('L')
        # self.image = ImageOps.grayscale(self.image)   # backup
        return self

    def _is_grayscale(self) -> bool:
        if len(self.image.getbands()) == 1:
            return True

    def binaryzation(self, threshold: int = 255 - 20) -> 'Img':
        """The grayscale greater than the threshold will be set to 255, otherwise set to 0"""
        if not self._is_grayscale():
            self.grayscale()
        self.image = self.image.point(lambda p: p > threshold and 255)
        return self

    def invert(self) -> 'Img':
        self.image = ImageOps.invert(self.image)
        return self

    def rotate(self, angle: float, expand: bool = True) -> 'Img':
        self.image = self.image.rotate(angle, expand=expand)
        return self

    def crop(self, rect: Rect) -> 'Img':
        self.image = self.image.crop(rect.to_ltrb())
        return self

    def dpi(self, dpi: int) -> 'Img':
        self.image.info["dpi"] = (dpi, dpi)
        return self

    def _get_save_params(self, **kwargs) -> dict:
        params = super()._get_save_params(**kwargs)
        if (not params.get('fp')) and hasattr(self, 'fp'):
            params['fp'] = self.fp
        if dpi := self.image.info.get('dpi'):
            params['dpi'] = dpi
        return params


class ImgDraw(Img):
    @staticmethod
    def _get_font_path() -> Path:
        has_font = True
        if is_sys(is_sys.LINUX):
            path = Path('/usr/share/fonts/truetype/wqy/wqy-zenhei.ttc')
        elif is_sys(is_sys.WINDOWS):
            path = Path(r'C:\Windows\Fonts\msyh.ttc')
        else:
            has_font = False
        if not has_font or not path.exists():
            mes = ('Font file not found, please use the class attribute "FONT_PATH" to designate another font path, '
                   'or try to use "sudo apt install -y ttf-wqy-zenhei" to install it on Linux.')
            raise FileNotFoundError(mes)
        return path
    FONT_PATH = _get_font_path()

    def __init__(self, image: Image.Image | np.ndarray | bytes | str | Path | io.BytesIO,
                 from_opencv: bool = False):
        super().__init__(image, from_opencv)
        self.draw = ImageDraw.Draw(self.image)

    @staticmethod
    def rgb_to_l(rgb_color: tuple[int, int, int] | list[int]) -> int:
        red, green, blue = rgb_color
        grayscale = 0.2989 * red + 0.5870 * green + 0.1140 * blue
        return round(grayscale)

    @staticmethod
    def bounding_rect(polygon: list | tuple | Rect) -> Rect:
        """find the smallest rectangle that can cover a polygon"""
        if isinstance(polygon, Rect):
            return polygon
        x_min = min(x for x, y in polygon)
        y_min = min(y for x, y in polygon)
        x_max = max(x for x, y in polygon)
        y_max = max(y for x, y in polygon)
        return Rect((x_min, y_min, x_max, y_max), rect_format=Rect.LTRB)

    def _get_color(self, outline: Any, box: tuple[int, int, int, int] | list[int]) -> Any:
        if not outline:
            region = self.image.crop(box)
            colors = region.getdata()
            avg_red = sum([color[0] for color in colors]) // len(colors)
            avg_green = sum([color[1] for color in colors]) // len(colors)
            avg_blue = sum([color[2] for color in colors]) // len(colors)

            grayscale = self.rgb_to_l((avg_red, avg_green, avg_blue))
            if grayscale < 128:
                return 'white'
            return 'black'
        return outline

    def rectangles(self, rects: list[Rect], outline: Any = None, width: float = 1) -> 'ImgDraw':
        for rect in rects:
            got_outline = self._get_color(outline, rect.to_ltrb())
            self.draw.rectangle(rect.to_ltrb(), outline=got_outline, width=width)
        return self

    def polygons(self, polygons: list, outline: Any = None, width: float = 1) -> 'ImgDraw':
        for polygon in polygons:
            got_outline = self._get_color(outline, self.bounding_rect(polygon).to_ltrb())
            self.draw.polygon(recursive_convert(polygon, to=tuple), outline=got_outline, width=width)
        return self

    def text(self, text: str, left_top_point: tuple[float, float], font_size: int = 10, fill: Any = None) -> 'ImgDraw':
        font = ImageFont.truetype(str(self.FONT_PATH), font_size)
        got_fill = self._get_color(fill, self.draw.textbbox(left_top_point, text, font=font))
        self.draw.text(left_top_point, text, font=font, fill=got_fill)
        return self

    def text_in_rect(self, text: str, rect: Rect, fill: Any = None) -> 'ImgDraw':
        def _get_font() -> ImageFont.FreeTypeFont:
            height, width = bottom - top, right - left
            font = ImageFont.truetype(str(self.FONT_PATH), size=int(height))
            while True:
                bbox = self.draw.textbbox(left_top_point, text, font=font)
                bbox_left, bbox_top, bbox_right, bbox_bottom = bbox
                if bbox_right > right or bbox_bottom > bottom:
                    height -= 1
                    font = ImageFont.truetype(str(self.FONT_PATH), size=int(height))
                else:
                    return font

        left, top, right, bottom = rect.to_ltrb()
        left_top_point = (left, top)
        text_font = _get_font()
        got_fill = self._get_color(fill, self.draw.textbbox(left_top_point, text, font=text_font))
        self.draw.text(left_top_point, text, font=text_font, fill=got_fill)
        return self



import datetime
import errno
import multiprocessing
import queue
import socket
import sys
import threading
import time
import traceback
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path
from typing import Any, Callable, Union
import requests
from psplpy.other_utils import is_sys
from psplpy.serialization_utils import Serializer


def _find(func: Callable, try_ports: list[int] = None, exclude_ports: list[int] = None,
          try_range: tuple[int, int] = None) -> int | None:
    ports = (try_ports or []) + list(range(*try_range))
    exclude_ports = exclude_ports or []
    for port in ports:
        if port not in exclude_ports:
            try:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    result = func(s, port)
                    if result is not None:
                        return result
            except socket.error:
                continue


def find_running_port(host: str = '127.0.0.1', try_ports: list[int] = None, exclude_ports: list[int] = None,
                      try_range: tuple[int, int] = None, timeout: float = 0.5) -> int | None:
    def _test(s: socket.socket, port: int):
        s.settimeout(timeout)
        result = s.connect_ex((host, port))
        if result == 0:
            return port

    return _find(_test, try_ports, exclude_ports, try_range)


def find_free_port(host: str = '127.0.0.1', try_ports: list[int] = None, exclude_ports: list[int] = None,
                   try_range: tuple[int, int] = (1024, 65536)) -> int | None:
    def _test(s: socket.socket, port: int):
        s.bind((host, port))
        return port

    return _find(_test, try_ports, exclude_ports, try_range)


class ClientSocket:
    def __init__(self, host: str = '127.0.0.1', port: int = 12345, client_socket: socket.socket = None,
                 client_host: str = None, client_port: int = 12345):
        self.host = host
        self.port = port
        self._serializer = Serializer()
        self.client_host = client_host
        self.client_port = client_port
        self.client_socket = client_socket

        self._length_bytes = 5

        if self.client_socket:
            if is_sys(is_sys.WINDOWS):
                self.client_socket.setblocking(True)
            self.client_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 32 * 1024 * 1024)

    def connect(self) -> None:
        self.client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.client_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 32 * 1024 * 1024)
        # use the certain address of the client to connect the server
        if self.client_host and self.client_port:
            self.client_socket.bind((self.client_host, self.client_port))
        self.client_socket.connect((self.host, self.port))

    def _get_length(self, data: bytes) -> bytes:
        # for 5 bytes unsigned int, the max data length is 2**40 - 1, namely about 1tb
        bytes_result = len(data).to_bytes(self._length_bytes, byteorder='big')
        return bytes_result

    def _recv_length(self) -> int:
        byte_result = self.client_socket.recv(self._length_bytes)
        return int.from_bytes(byte_result, byteorder='big')

    def send(self, data: bytes):
        return self.client_socket.sendall(self._get_length(data) + data)

    def recv(self) -> bytes:
        length = self._recv_length()
        data = bytearray()
        while len(data) < length:
            new_data = self.client_socket.recv(length - len(data))
            data += new_data
            if len(new_data) == 0:  # avoid the sender ended by accident, lead to get into the infinite loop
                return bytes(data)
        return bytes(data)

    def recvf(self, output_path: str | Path, bufsize: int = 1024 * 1024 * 16) -> None:
        with open(output_path, 'wb') as f:
            while True:
                data = self.client_socket.recv(bufsize)
                if not data:
                    break
                f.write(data)

    def sendf(self, input_path: str | Path, bufsize: int = 1024 * 1024 * 16) -> None:
        with open(input_path, 'rb') as f:
            while True:
                data = f.read(bufsize)
                if not data:
                    break
                self.client_socket.send(data)

    def send_pickle(self, data: Any) -> None:
        return self.send(self._serializer.dump_pickle(data))

    def recv_pickle(self) -> Any | None:
        if data := self.recv():
            return self._serializer.load_pickle(data)

    def close(self) -> None:
        return self.client_socket.close()


class ServerSocket:
    def __init__(self, host: str = '127.0.0.1', port: int = 12345, backlog: int = 64):
        self.host = host
        self.port = port
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.bind((self.host, self.port))
        self.server_socket.listen(backlog)
        self.server_socket.setblocking(False)

    def accept(self) -> tuple[ClientSocket, Any] | tuple[None, None]:
        while True:
            try:
                client_socket, addr = self.server_socket.accept()
                return ClientSocket(client_socket=client_socket), addr
            except socket.error as e:
                if e.errno == errno.EWOULDBLOCK or e.errno == errno.EAGAIN:
                    time.sleep(0.01)
                    continue
                elif e.errno == errno.EBADF:
                    return None, None
                elif is_sys(is_sys.WINDOWS):
                    if e.errno == errno.WSAENOTSOCK:
                        return None, None
                raise e

    def handle(self, handler: Callable, *args, **kwargs) -> None:
        def _handle():
            while True:
                client_socket, addr = self.accept()
                if not (client_socket and addr):
                    break
                handle_thread = threading.Thread(target=handler, args=(client_socket, addr, *args), kwargs=kwargs)
                handle_thread.daemon = True
                handle_thread.start()

        threading.Thread(target=_handle).start()

    def close(self) -> None:
        self.server_socket.close()


class MpHttpError(Exception):
    def __init__(self, *args, traceback_info: str = ''):
        super().__init__(*args)
        self.traceback_info = traceback_info

    def __str__(self):
        return self.traceback_info
    __repr__ = __str__


class MpHttpServer:
    SUBMIT = '/submit'
    FETCH = '/fetch'
    GET_LOAD = '/get_load'
    GET_PROGRESS = '/get_progress'
    _POLL_INTERVAL = 0.1

    def __init__(self, host: str = '0.0.0.0', port: int = 80, workers: int = 1, show_info: bool = True,
                 max_fetch_timeout: float = 3600, result_timeout: float = 3600):
        self.host, self.port, self.workers, self.show_info = host, port, workers, show_info
        self.max_fetch_timeout, self.result_timeout = max_fetch_timeout, result_timeout
        self._task_id = 0
        self._lock = multiprocessing.Lock()  # Do not use threading.Lock, can't be pickled on Windows
        self._s = Serializer()
        self._result_dict: dict[int, MpHttpServer._TaskResult] = {}
        self._req_que = multiprocessing.Queue()
        self._result_que = multiprocessing.Queue()
        self._closed_flag = multiprocessing.Value('b', False)
        self._load = multiprocessing.Value('i', 0)

    class _SubTask:
        def __init__(self, task_id: int, sub_task_id: int, data: Any):
            self.task_id = task_id
            self.sub_task_id = sub_task_id
            self.data = data
            self.result = None

    class _TaskResult:
        def __init__(self, task_num: int):
            self.task_num = task_num
            self.finished_task_num = 0
            self.finished_task_list = [None] * task_num  # [None] * 0 = []
            self.is_finished = threading.Event()
            self.finished_time = None
            if self.task_num == 0:
                self._set_finished()

        def _set_finished(self):
            self.is_finished.set()
            self.finished_time = time.time()

        def add_finished_subtask(self, sub_task: 'MpHttpServer._SubTask') -> None:
            self.finished_task_list[sub_task.sub_task_id] = sub_task.result
            self.finished_task_num += 1
            if self.finished_task_num == self.task_num:
                self._set_finished()

    class _RequestHandler(BaseHTTPRequestHandler):
        def __init__(self, s: 'MpHttpServer', *args, **kwargs):
            self.s = s
            super().__init__(*args, **kwargs)

        def log_request(self, code="-", size="-"):
            pass

        def log_message(self, format, *args):
            if self.s.show_info:
                super().log_message(format, *args)

        def _put_data(self, data: Any) -> int:
            sub_task_id = 0
            with self.s._lock:
                task_id = self.s._task_id
                self.s._task_id += 1
            for sub_data in data:
                self.s._req_que.put(self.s._SubTask(task_id, sub_task_id, sub_data))
                sub_task_id += 1
            self.s._result_dict[task_id] = self.s._TaskResult(len(data))
            return task_id

        def _fetch_result(self, task_id: int, timeout: float) -> list[Any] | TimeoutError | KeyError:
            timeout = timeout if timeout <= self.s.max_fetch_timeout else self.s.max_fetch_timeout
            if self.s._result_dict.get(task_id):
                if self.s._result_dict[task_id].is_finished.wait(timeout):
                    return self.s._result_dict.pop(task_id).finished_task_list
                return TimeoutError(f'Fetch task {task_id} result timeout after {timeout}s')
            return KeyError(f'Task {task_id} result not exist or expired')

        def do_POST(self):
            content_length = int(self.headers['Content-Length'])
            data = []
            if content_length:
                post_data = self.rfile.read(content_length)
                data = self.s._s.load_pickle(post_data)

            code = 200
            if self.path == MpHttpServer.GET_LOAD:
                result = self.s._load.value / self.s.workers
            elif self.path == MpHttpServer.GET_PROGRESS:
                task = self.s._result_dict.get(data)
                result = (0, 0)
                if task:
                    result = (task.finished_task_num, task.task_num)
            elif self.path == MpHttpServer.SUBMIT:
                result = self._put_data(data)
            elif self.path == MpHttpServer.FETCH:
                result = self._fetch_result(**data)
            else:
                code, result = 404, None

            response = self.s._s.dump_pickle(result)
            self.log_message('"%s" %s %s %s', self.requestline, str(code),
                             str(len(response)), str(self.client_address))
            self.send_response(code)
            self.send_header('Content-Type', 'application/octet-stream')
            self.send_header('Content-Length', str(len(response)))
            self.end_headers()
            self.wfile.write(response)

    @staticmethod
    def _get_subtask(func: Callable):
        def wrapper(self, que: multiprocessing.Queue, *args, **kwargs) -> None:
            while True:
                try:
                    sub_task: MpHttpServer._SubTask = que.get(timeout=MpHttpServer._POLL_INTERVAL)
                except queue.Empty:
                    if self._closed_flag.value:
                        break
                    continue
                func(self, sub_task, *args, **kwargs)
        return wrapper

    @_get_subtask
    def _result_distribution(self, sub_task: Union[multiprocessing.Queue, 'MpHttpServer._SubTask']) -> None:
        self._result_dict[sub_task.task_id].add_finished_subtask(sub_task)

    def _cleanup_expired_results(self):
        last_cleanup_time = time.time()
        while True:
            if self._closed_flag.value:
                break
            if time.time() - last_cleanup_time > self.result_timeout / 10:
                expired_task_ids = []
                for task_id, task in self._result_dict.items():
                    if task.is_finished.is_set() and time.time() - task.finished_time > self.result_timeout:
                        expired_task_ids.append(task_id)
                for task_id in expired_task_ids:
                    self._result_dict.pop(task_id)
                    if self.show_info:
                        time_str = datetime.datetime.now().strftime("[%d/%b/%Y %H:%M:%S]")
                        sys.stderr.write(f'Warning - {time_str} - Result of task {task_id} has expired\n')
                last_cleanup_time = time.time()
            else:
                time.sleep(self._POLL_INTERVAL)

    def init(self) -> None: ...

    def main_loop(self, data: Any) -> Any:
        return data

    def _main_process(self, req_que: multiprocessing.Queue, result_que: multiprocessing.Queue) -> None:
        @MpHttpServer._get_subtask
        def _process(self, sub_task: Union[multiprocessing.Queue, 'MpHttpServer._SubTask']) -> None:
            with self._load.get_lock():
                self._load.value += 1
            try:
                result = self.main_loop(sub_task.data)
                sub_task.result = result
            except Exception:
                sub_task.result = MpHttpError(traceback_info=traceback.format_exc())
            finally:
                del sub_task.data
                result_que.put(sub_task)
                with self._load.get_lock():
                    self._load.value -= 1

        self.init()
        _process(self, req_que)

    def run_server(self, new_thread: bool = False) -> 'MpHttpServer':
        threading.Thread(target=self._result_distribution, args=(self._result_que,)).start()
        threading.Thread(target=self._cleanup_expired_results).start()
        for _ in range(self.workers):
            multiprocessing.Process(target=self._main_process, args=(self._req_que, self._result_que)).start()
        self._httpd = ThreadingHTTPServer((self.host, self.port),
                                          lambda *args, **kwargs: self._RequestHandler(self, *args, **kwargs))
        if self.show_info:
            sys.stderr.write(f"Starting server on ({self.host}, {self.port})...\n")
        if new_thread:
            threading.Thread(target=self._httpd.serve_forever, kwargs={'poll_interval': self._POLL_INTERVAL}).start()
        else:
            self._httpd.serve_forever(poll_interval=self._POLL_INTERVAL)
        return self

    def close_server(self) -> None:
        self._closed_flag.value = True
        self._httpd.shutdown()
        self._httpd.server_close()


class MpHttpClient:
    def __init__(self, host: str = '127.0.0.1', port: int = 80, server: MpHttpServer = None):
        self.host, self.port = host, port
        if server:
            self.host, self.port = server.host if server.host != '0.0.0.0' else '127.0.0.1', server.port
        self._s = Serializer()

    def _req(self, data: Any, path: str) -> Any:
        pickled_data = self._s.dump_pickle(data)
        resp = requests.post(f'http://{self.host}:{self.port}{path}', data=pickled_data)
        return self._s.load_pickle(resp.content)

    def submit(self, data_list: list | tuple) -> int:
        """Getting the task_id for fetching data from the server"""
        return self._req(data_list, MpHttpServer.SUBMIT)

    def fetch(self, task_id: int, timeout: float = 3600) -> list[Any]:
        """The max timeout time depends on the setting of the server"""
        result = self._req({'task_id': task_id, 'timeout': timeout}, MpHttpServer.FETCH)
        if isinstance(result, (TimeoutError, KeyError)):
            raise result
        return result

    def batch(self, data_list: list | tuple, timeout: float = 3600) -> list[Any]:
        return self.fetch(self.submit(data_list), timeout)

    def get(self, data: Any, timeout: float = 3600) -> Any:
        return self.batch([data], timeout)[0]

    def get_load(self) -> float:
        return self._req(None, MpHttpServer.GET_LOAD)

    def get_progress(self, task_id: int) -> tuple[int, int]:
        """return the tuple (finished_task_num, total_task_num)"""
        return self._req(task_id, MpHttpServer.GET_PROGRESS)



import datetime
import math
import multiprocessing
import platform
import re
import sys
import time
from abc import ABC, abstractmethod
from typing import Any, Type, Union
import json
import os
from pathlib import Path
import cv2
import numpy as np
from PIL import ImageGrab, Image


class FunctionClass(ABC):
    def __new__(cls, *args, **kwargs) -> Any:
        instance = super().__new__(cls)
        instance._init()
        return instance(*args, **kwargs)

    def _init(self) -> None: ...

    @abstractmethod
    def __call__(self, *args, **kwargs): ...


class is_sys(FunctionClass):
    WINDOWS = 'Windows'
    LINUX = 'Linux'
    MACOS = 'Darwin'

    def __new__(cls, os_name: str) -> bool:
        return super().__new__(cls, os_name)

    def __call__(self, os_name: str) -> bool:
        this_os_name = platform.system()
        if os_name == this_os_name:
            return True
        return False


def recursive_convert(data: list | tuple, to: Type) -> tuple | list:
    """Recursively convert the lists and tuples are nested within each other to only tuples or lists"""
    if isinstance(data, (list, tuple)):
        return to(recursive_convert(item, to=to) for item in data)
    return data


def split_list(lst: list, n: int) -> list[list]:
    """Split a list into n equal length sublists in original order"""
    size = len(lst) // n
    remainder = len(lst) % n

    result_lists = []
    start = 0

    for i in range(n):
        if i < remainder:
            end = start + size + 1
        else:
            end = start + size

        result_lists.append(lst[start:end])
        start = end

    return result_lists


def get_key_from_value(dct: dict, value: Any, find_all: bool = False,
                       allow_null: bool = False) -> Union[Any, None]:
    keys = []
    for key, val in dct.items():
        if val == value:
            if not find_all:
                return key
            keys.append(key)
    if not keys:
        if allow_null:
            return None
        raise KeyError(f"No corresponding key exists")
    return keys


class get_env(FunctionClass):
    ENV_FILE = '.env'

    @staticmethod
    def _parse_key_value() -> None:
        result = {}
        lines = Path(get_env.ENV_FILE).read_text(encoding='utf-8').strip().split('\n')
        for line in lines:
            line = line.strip()
            if line and not line.startswith('#'):
                key_value = line.split('=', maxsplit=2)
                result[key_value[0].strip()] = key_value[1].strip()
        for key, value in result.items():
            os.environ[key] = value

    def _init(self) -> None:
        self._has_read = False

    def __new__(cls, env_name: str, reload: bool = False) -> Any:
        return super().__new__(cls, env_name, reload)

    def __call__(self, env_name: str, reload: bool = False) -> Any:
        if reload:
            self._parse_key_value()
        env = os.environ.get(env_name)
        try:
            return json.loads(env)
        except json.decoder.JSONDecodeError:
            return env
        except TypeError:
            if self._has_read:
                raise KeyError(f'Environment variable {env_name} not exist.')
            self._parse_key_value()
            self._has_read = True
            return self.__call__(env_name, reload)


class PerfCounter:
    US = 'us'
    MS = 'ms'
    S = 's'
    disabled = False

    def __init__(self, unit: str = MS, places: int = 4):
        if not self.disabled:
            self.unit = unit
            self.places = places

            self._start = time.perf_counter()

    def refresh(self) -> None:
        if not self.disabled:
            self._start = time.perf_counter()

    def elapsed(self) -> float:
        if not self.disabled:
            elapsed = time.perf_counter() - self._start
            if self.unit == self.MS:
                elapsed = elapsed * 1000
            elif self.unit == self.US:
                elapsed = elapsed * 1000 * 1000
            self.refresh()
            return elapsed
        return 0

    def show(self, tag: str = '') -> None:
        if not self.disabled:
            if tag:
                tag = tag + ': '
            print(f'{tag}{self.elapsed():.{self.places}f}{self.unit}')
            self.refresh()


class TimeoutChecker:
    instances = {}

    def __new__(cls, timeout_seconds: float, unique_id: Any = None):
        if unique_id in cls.instances:
            cls.instances[unique_id]['is_first_created'] = False
            return cls.instances[unique_id]['instance']
        else:
            instance = super().__new__(cls)
            cls.instances[unique_id] = {'instance': instance, 'is_first_created': True}
            return instance

    def __init__(self, timeout_seconds: float, unique_id: Any = None):
        """unique_id is required when two or more instances existing simultaneously, e.g. using threading"""
        self.unique_id = unique_id
        if self.instances[self.unique_id]['is_first_created']:
            self.timeout_seconds = timeout_seconds
            self.start_time = time.time()

    def _delete_instance(self):
        if self.unique_id in self.instances:
            del self.instances[self.unique_id]

    def ret_false(self) -> bool:
        current_time = time.time()
        if current_time - self.start_time > self.timeout_seconds:
            self._delete_instance()
            return False
        return True

    def raise_err(self, error_type: Type = None) -> bool:
        if not self.ret_false():
            if not error_type:
                message = (f'Start at {time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.start_time))}, '
                           f'after {self.timeout_seconds}s time out')
                raise TimeoutError(message)
            raise error_type()
        return True


class ScreenShotMaker:
    PNG = 'png'  # using PIL saves to png, the smallest size, but the longest time
    BMP = 'bmp'  # using PIL saves to bmp, the biggest size, but the shortest time
    PNG_OPENCV = 'png_opencv'  # using opencv saves to png, medium size and time

    def __init__(self, region: (int, int, int, int) = None, fps: float = 10, save_dir: str | Path = None,
                 show_info: bool = True, enable_warning: bool = True, savers_num: int | str = 'auto',
                 save_to: str = PNG_OPENCV) -> None:
        self._region, self._fps, self._show_info = region, fps, show_info
        self._enable_warning, self._save_to = enable_warning, save_to

        self._interval = 1 / self._fps
        self._savers_num = self._get_savers_num(savers_num)
        self._frame_counter, self._frame_queue = 0, multiprocessing.Queue()

        self._save_dir = Path(save_dir) if save_dir is not None else Path().cwd()
        self._save_dir = self._save_dir / f'{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S_%f")}'
        self._save_dir.mkdir(parents=True, exist_ok=True)

        self._start_savers()
        self._screenshot()

    def _get_savers_num(self, savers_num: int | str) -> int:
        if savers_num == 'auto':
            if self._save_to == self.PNG:
                savers_num = self._fps
            elif self._save_to == self.BMP:
                savers_num = math.ceil(self._fps / 10)
            elif self._save_to == self.PNG_OPENCV:
                savers_num = math.ceil(self._fps / 3)
        return savers_num

    def _warning(self, lag: float) -> None:
        if self._enable_warning and lag > self._interval:
            print(f'Warning - {datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S.%f")} - '
                  f'frame {self._frame_counter} - the lag is {lag:.3f}s')
        if self._enable_warning and self._frame_queue.qsize() > self._fps:
            print(f'Warning - {datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S.%f")} - '
                  f'frame {self._frame_counter} - the queue size is {self._frame_queue.qsize()}')

    @staticmethod
    def _timer(func):
        def wrapper(self, *args, **kwargs):
            t_start = time.perf_counter()
            while True:
                expected_time = t_start + self._frame_counter * self._interval
                current_time = time.perf_counter()
                lag = current_time - expected_time
                if current_time > expected_time:
                    self._warning(lag)
                    result = func(self, *args, **kwargs)
                    self._frame_queue.put({'time': current_time, 'result': result, 'counter': self._frame_counter})
                    self._frame_counter += 1
                else:
                    time.sleep(0.001)
        return wrapper

    @_timer
    def _screenshot(self):
        return ImageGrab.grab(self._region)

    def _save(self, image: Image.Image, path: str):
        if self._save_to == self.PNG_OPENCV:
            image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
            cv2.imwrite(f'{path}.png', image)
        elif self._save_to == self.PNG:
            image.save(f'{path}.png')
        elif self._save_to == self.BMP:
            image.save(f'{path}.bmp')
        else:
            raise AssertionError

    def _saver(self, frame_queue: multiprocessing.Queue):
        while True:
            if not frame_queue.empty():
                frame = frame_queue.get()
                self._save(frame['result'], str(self._save_dir / f'{frame["time"]:.2f}'))
                if self._show_info:
                    print(f'At {frame["time"]:.3f}, save frame {frame["counter"]}')
            else:
                time.sleep(0.01)

    def _start_savers(self):
        for _ in range(self._savers_num):
            multiprocessing.Process(target=self._saver, args=(self._frame_queue,)).start()


def overlay_print(s: str) -> None:
    sys.stdout.write(f'\r{s}')  # \r means go back to the beginning of the line
    sys.stdout.flush()


def progress_bar(progress: float, show_str: str = '', bar_length: int = 40,
                 finished_chr: str = '=', unfinished_chr: str = '-',
                 left_border_chr: str = '[', right_border_chr: str = ']',
                 progress_precision: int = 2) -> None:
    if progress > 1:
        progress = 1
    filled_length = int(progress * bar_length)
    bar = finished_chr * filled_length + unfinished_chr * (bar_length - filled_length)
    show_str = show_str if show_str else f'{progress * 100:.{progress_precision}f}%'
    overlay_print(f'{left_border_chr}{bar}{right_border_chr} {show_str}')


def limited_input(str_list: [list | tuple | set] = None, regex_list: [list | tuple | set] = None,
                  func_list: [list | tuple | set] = None, input_processing_func=None,
                  print_str: str = '', error_tip: str = 'Invalid input, please re-enter.') -> str:
    result_str = ''
    while True:
        input_str = input(print_str + '\n')
        if str_list:
            if input_str in str_list:
                result_str = input_str
        if not result_str and regex_list:
            for regex in regex_list:
                if re.match(regex, input_str):
                    result_str = input_str
                    break
        if not result_str and func_list:
            for func in func_list:
                if func(input_str):
                    result_str = input_str
                    break
        if result_str:
            if input_processing_func:
                return input_processing_func(result_str)
            return result_str
        print(error_tip)


class VideoToFrames:
    def __init__(self, video_path: str | Path, save_dir: str | Path = '', frame_interval: int = 1,
                 show_progress: bool = True, workers: int = 5) -> None:
        self.video_path = video_path
        self.save_dir = Path(save_dir) / Path(video_path).stem
        self.save_dir.mkdir(parents=True, exist_ok=True)
        self.frame_interval = frame_interval
        self.show_progress = show_progress
        self._progress = multiprocessing.Value('i', 0)
        self.workers = workers

        cap = cv2.VideoCapture(self.video_path)
        self.total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        self.num_digits = len(str(self.total_frames))  # 计算帧数的填充位数

        indices = self._split_number_into_parts(self.total_frames, self.workers)
        for i in range(self.workers):
            multiprocessing.Process(target=self._save_process, args=(indices[i], indices[i + 1])).start()

    @staticmethod
    def _split_number_into_parts(number: int, parts: int) -> list[int]:
        if parts <= 0 or number <= 0 or not isinstance(parts, int) or not isinstance(number, int):
            raise ValueError("Number and parts should be int and greater than 0.")
        part_size = number // parts
        return [i * part_size for i in range(parts)] + [number]

    def _save_process(self, start_frame: int, stop_frames: int) -> None:
        cap = cv2.VideoCapture(self.video_path)
        while start_frame < stop_frames:
            ret, frame = cap.read()
            if start_frame % self.frame_interval == 0:
                frame_filename = self.save_dir / f"frame_{start_frame:0{self.num_digits}d}.png"
                cv2.imwrite(str(frame_filename), frame)

            start_frame += 1
            with self._progress.get_lock():
                self._progress.value += 1
                if self.show_progress:
                    progress_bar(self._progress.value / self.total_frames,
                                 show_str=f'{self._progress.value}/{self.total_frames}')
        cap.release()



import tests.test_middleware_utils
import tests.test_serialization_utils
import tests.test_concurrency_utils
import tests.test_network_utils
import tests.test_file_utils
import tests.test_image_utils
import tests.test_other_utils
import tests.test_dynamic_compose
import tests.test_ocr_utils
import tests.test_img_det


def global_test():
    tests.test_middleware_utils.tests()
    tests.test_serialization_utils.tests()
    tests.test_concurrency_utils.tests()
    tests.test_network_utils.tests()
    tests.test_file_utils.tests()
    tests.test_image_utils.tests()
    tests.test_other_utils.tests()
    tests.test_dynamic_compose.tests()
    tests.test_ocr_utils.tests()
    tests.test_img_det.tests()


if __name__ == '__main__':
    global_test()



import os
import shutil
import sys
from pathlib import Path


class UploadToPypi:
    def __init__(self, project_dir: str | Path, username: str = '__token__', passwd: str = None,
                 upload_to_pypi_test: bool = False, emit_tips: bool = True, python_path: str = None):
        self.project_dir = Path(project_dir).resolve()
        self.username = username
        self.passwd = passwd
        self.upload_to_pypi_test = upload_to_pypi_test
        self.emit_tips = emit_tips
        self.python_path = python_path or sys.executable

        os.chdir(self.project_dir)

    def build(self):
        try:
            shutil.rmtree(self.project_dir / 'dist')    # delete all old dist versions
        except FileNotFoundError:
            if self.emit_tips:
                print(f"Dist doesn't exist or failed to delete.")
        os.system(f'{self.python_path} -m build')

    def upload(self):
        command = f'{self.python_path} -m twine upload dist/* -u {self.username} -p {self.passwd}'
        if self.upload_to_pypi_test:
            command += ' --repository testpypi'
        os.system(command)

    def build_and_upload(self):
        self.build()
        self.upload()



import threading
from typing import Any, Callable
import pika
import requests
from requests.auth import HTTPBasicAuth
from psplpy.serialization_utils import Serializer, CompressSerializer


class _RabbitmqManagement:
    def __init__(self, host: str, management_port: int, user: str, pw: str):
        self.api_url = f'http://{host}:{management_port}/api'
        self.auth = HTTPBasicAuth(user, pw)

    def _vhost(self, vhost_name: str, method: str) -> str:
        url = f'{self.api_url}/vhosts/{vhost_name}'
        response = getattr(requests, method)(url, auth=self.auth)
        if response.status_code not in [201, 204]:
            raise AssertionError(f'Failed to {method} vhost: {vhost_name}, status code: {response.status_code}')
        return vhost_name

    def create_vhost(self, vhost_name: str) -> str:
        return self._vhost(vhost_name, 'put')

    def delete_vhost(self, vhost_name: str) -> str:
        return self._vhost(vhost_name, 'delete')

    def list_vhosts(self) -> list[str]:
        url = f'{self.api_url}/vhosts'
        response = requests.get(url, auth=self.auth)
        if response.status_code == 200:
            vhosts = response.json()
            return [vhost['name'] for vhost in vhosts]
        raise AssertionError(f'Failed to list vhosts, status code: {response.status_code}')


class Rabbitmq:
    HOST = 'localhost'
    PORT = 5672
    MANAGEMENT_PORT = 15672
    USER = 'guest'
    PW = 'guest'
    VIRTUAL_HOST = '/'

    JSON = 'j'
    PICKLE = 'p'
    AUTO = CompressSerializer.AUTO

    def __init__(self, host: str = None, port: int = None, user: str = None, pw: str = None, virtual_host: str = None,
                 management_port: int = None, serializer: str = PICKLE, compress: bool | str = False,
                 compress_threshold: int = 1024 * 128, **kwargs):
        self.host = host or self.HOST
        self.port = port or self.PORT
        self.user = user or self.USER
        self.pw = pw or self.PW
        self.virtual_host = virtual_host or self.VIRTUAL_HOST
        self.management_port = management_port or self.MANAGEMENT_PORT
        self.serializer, self.compress, self.compress_threshold = serializer, compress, compress_threshold
        self._serializer = CompressSerializer(compress=self.compress, threshold=self.compress_threshold)
        self.management = _RabbitmqManagement(self.host, self.management_port, self.user, self.pw)

        credentials = pika.PlainCredentials(self.user, self.pw)
        parameters = pika.ConnectionParameters(self.host, self.port, self.virtual_host, credentials, **kwargs)
        self.connection = pika.BlockingConnection(parameters)
        self.channel = self.connection.channel()

        self._stop_consuming = False
        self._suppress_error = False

    def _serialize(self, data: Any) -> bytes:
        if self.serializer == self.PICKLE:
            return self.PICKLE.encode() + self._serializer.dump_pickle(data)
        else:
            raise ValueError(f'Unsupported serializer: {self.serializer}')

    def _deserialize(self, data: bytes) -> Any:
        serializer = chr(data[0])
        data = data[1:]
        if serializer == self.PICKLE:
            return self._serializer.load_pickle(data)
        else:
            raise ValueError(f'Unsupported serializer: {self.serializer}')

    @staticmethod
    def _default_callback(ch, method, properties, body) -> None:
        print(body)

    def _callback(self, ch, method, properties, body) -> None:
        body = self._deserialize(body)
        self.callback(ch, method, properties, body)
        if self._stop_consuming:
            ch.stop_consuming()

    def send_init(self, exchange: str, routing_keys: list[str]) -> None:
        self.exchange = exchange
        self.routing_keys = routing_keys
        self.channel.exchange_declare(exchange=self.exchange, exchange_type='topic')

    def recv_init(self, exchange: str, binding_keys: list[str], callback: Callable = None) -> None:
        self.send_init(exchange, routing_keys=binding_keys)
        self.callback = callback or self._default_callback
        queue = self.channel.queue_declare('', exclusive=True)
        self.queue_name = queue.method.queue
        for routing_key in self.routing_keys:
            self.channel.queue_bind(exchange=self.exchange, queue=self.queue_name, routing_key=routing_key)
        self.channel.basic_consume(queue=self.queue_name, on_message_callback=self._callback, auto_ack=True)

    def basic_publish(self, body: Any) -> None:
        for routing_key in self.routing_keys:
            self.channel.basic_publish(exchange=self.exchange, routing_key=routing_key, body=self._serialize(body))

    def _start_consuming(self) -> None:
        try:
            self.channel.start_consuming()
        except Exception as e:
            if not self._suppress_error:
                raise e

    def start_consuming(self) -> threading.Thread:
        t = threading.Thread(target=self._start_consuming)
        t.start()
        return t

    def stop_consuming(self):
        self._stop_consuming = True

    def close(self, suppress_error: bool = False) -> None:
        self._suppress_error = suppress_error
        try:
            self.connection.close()
        except Exception as e:
            if not self._suppress_error:
                raise e



from tests.__init__ import *
from psplpy.dynamic_compose import DynamicCompose

dc_rc_dir = rc_dir / 'dynamic_compose'
dc_compose_dir = dc_rc_dir / '.compose'
compose_dumped_file = dc_compose_dir / 'compose-dumped.yml'
dockerfile_dumped_file = dc_compose_dir / 'Dockerfile-dumped'


def lines_rstrip(text: str) -> str:
    lines = text.splitlines()
    stripped_lines = [line.rstrip() for line in lines]
    result = '\n'.join(stripped_lines)
    return result


def tests():
    DynamicCompose.CWD = dc_rc_dir
    dc = DynamicCompose()
    dc.arg['GPU'] = '''    
    deploy:
      resources:
        reservations:
          devices:
            - driver: nvidia
              count: all
              capabilities: [gpu]'''
    dc.arg['IMAGE'] = 'python:3.10'
    print(dc.env)
    print(dc.arg)
    dc.compose_dir = dc_compose_dir

    dc.format_compose()
    dc.format_dockerfile()
    dc.dump()

    assert compose_dumped_file.read_text().strip() == lines_rstrip(dc.compose_file.read_text().strip())
    # dc.compose_file.unlink()
    assert dockerfile_dumped_file.read_text().strip() == lines_rstrip(dc.dockerfile_file.read_text().strip())
    # dc.dockerfile_file.unlink()

    dc.up()


if __name__ == '__main__':
    tests()



import random
from tests.__init__ import *
from psplpy.serialization_utils import *
from psplpy.other_utils import PerfCounter


def test_bit_manipulator():
    bm0, bm1, bm256 = BitManipulator(0, max_bits=8), BitManipulator(1), BitManipulator(256)
    # test __len__
    assert len(bm0) == 0, len(bm256) == 9
    # test __setitem__
    bm0[7] = 1
    bm256[15] = 1
    assert bm0.bit_value == 0b10000000 and bm256.bit_value == 0b10000001_00000000
    bm0[1:8:2] = '1001'
    assert str(bm0) == '10000010', str(bm0)
    try:
        bm0[9:10] = '11'
    except ValueError as e:
        print(e)
    else:
        assert False
    try:
        bm0[0] = '11'
    except ValueError as e:
        print(e)
    else:
        assert False
    bm1[9:11] = '11'    # out of the original bit range
    assert bm1.bit_value == 0b110_00000001
    # test max_bits
    try:
        bm0.set_bit(8, 1)
    except ValueError as e:
        print(e)
    else:
        assert False
    try:
        bm0.set_bit(8, 0)
    except ValueError as e:
        assert False
    # test _index_check, __getitem__
    try:
        _ = bm0[-9]
    except IndexError as e:
        print(e)
    else:
        assert False
    assert str(bm0) == '10000010'
    assert bm0[128:256] == '0' * 128    # ability to slice out of range
    assert bm0[1024] == '0'
    assert bm0[7] == '1'
    assert bm0[4:8] == '0001'
    # test to_bytes
    b = bm0.to_bytes()
    assert b == b'\x82'
    assert bin(int.from_bytes(b, byteorder='big'))[2:] == '10000010'
    # test __format__
    assert bm0.__format__('o') == '202'
    assert bm0.__format__('x') == '82'
    assert bm0.__format__('d') == '130'
    # test __iter__
    assert ''.join([b for b in bm0]) == '10000010'[::-1]
    # test __lt__ __ne__ __ge__
    assert bm0 < bm256
    assert bm0 != bm256
    assert bm0 >= BitManipulator(bm0)
    # test fixed_bits
    bm = BitManipulator(0, fixed_bits=14)
    try:
        bm.set_bit(16, 1)
    except ValueError as e:
        print(e)
    else:
        assert False
    try:
        bm.set_bit(16, 0)
    except ValueError as e:
        assert False
    assert bm.to_bytes() == b'\x00\x00'
    assert bm.to_bytes(3) == b'\x00\x00\x00'
    assert BitManipulator(0).to_bytes() == b''
    assert str(bm) == '0' * 14, str(bm)
    assert bm.__format__('x') == '0'


bench_data = {}
rand_round = 10000
for i in range(rand_round):
    bench_data[str(random.randint(0, rand_round))] = random.uniform(0, rand_round)
python_data = {1: '100', 2: 200, 3: ['你好', [3.14, None, False]]}
json_str = '{"1":"100","2":200,"3":["你好",[3.14,null,false]]}'
pickle_bytes = (b'\x80\x04\x95/\x00\x00\x00\x00\x00\x00\x00}\x94(K\x01\x8c\x03100\x94K\x02K\xc8K\x03]\x94('
                b'\x8c\x06\xe4\xbd\xa0\xe5\xa5\xbd\x94]\x94(G@\t\x1e\xb8Q\xeb\x85\x1fN\x89eeu.')
yaml_str = "1: '100'\n2: 200\n3:\n- 你好\n- - 3.14\n  - null\n  - false\n"


def test_compress_serializer():
    zlib_serializer = CompressSerializer(threshold=1)
    lzma_serializer = CompressSerializer(threshold=1, compress_lib=Compressor.LZMA)

    p = PerfCounter()
    serialized_data = zlib_serializer.dump_pickle(bench_data)
    print(f'len: {len(serialized_data)},\t elapsed: {p.elapsed():.4f}ms,\t zlib_pickle')
    serialized_data = lzma_serializer.dump_pickle(bench_data)
    print(f'len: {len(serialized_data)},\t elapsed: {p.elapsed():.4f}ms,\t lzma_pickle')

    for dump, load, load_kwargs in [('dump_pickle', 'load_pickle', {}), ('dump_yaml', 'load_yaml', {}),
                                    ('dump_json', 'load_json', {'trans_key_to_num': True})]:
        dumps_data = getattr(zlib_serializer, dump)(python_data)
        assert isinstance(dumps_data, bytes), dump
        loads_data = getattr(zlib_serializer, load)(dumps_data, **load_kwargs)
        assert loads_data == python_data, f'{load}, {loads_data}'

    dumps_data = zlib_serializer.dump_pickle(python_data)
    assert zlib_serializer.load(dumps_data) == python_data
    dumps_data = zlib_serializer.dump_json(python_data)
    assert zlib_serializer.load(dumps_data, trans_key_to_num=True) == python_data


    compress_serializer = CompressSerializer(path=tmp_file, data_type=dict, compress=True)
    uncompress_serializer = CompressSerializer(path=tmp_file, data_type=dict, compress=False)

    for serializer in [compress_serializer, uncompress_serializer]:
        try:
            print(f'compress: {serializer.compress}')
            loads_data = serializer.load_json()
            assert loads_data == dict(), loads_data
            loads_data = serializer.load_json()
            assert loads_data == dict(), loads_data

            dumps_data = serializer.dump_json(python_data)
            assert isinstance(dumps_data, bytes)
            loads_data = serializer.load_json(trans_key_to_num=True)
            assert loads_data == python_data, loads_data

            dumps_data = serializer.dump_yaml(python_data)
            assert isinstance(dumps_data, bytes)
            loads_data = serializer.load_yaml()
            assert loads_data == python_data, loads_data

            dumps_data = serializer.dump_pickle(python_data)
            assert isinstance(dumps_data, bytes)
            loads_data = serializer.load_pickle()
            assert loads_data == python_data, loads_data
        finally:
            tmp_file.unlink(missing_ok=True)


def test_serializer():
    serializer = Serializer()
    p = PerfCounter()
    serialized_data = serializer.dump_pickle(bench_data)
    print(f'len: {len(serialized_data)},\t elapsed: {p.elapsed():.4f}ms,\t dumps_pickle')
    serialized_data = serializer.dump_json(bench_data)
    print(f'len: {len(serialized_data)},\t elapsed: {p.elapsed():.4f}ms,\t dumps_json')
    serialized_data = serializer.dump_yaml(bench_data)
    print(f'len: {len(serialized_data)},\t elapsed: {p.elapsed():.4f}ms,\t dumps_yaml')

    dumps_data = serializer.dump_json(python_data, ensure_ascii=False)
    loads_data = serializer.load_json(dumps_data, trans_key_to_num=True)
    assert dumps_data == json_str, dumps_data
    assert loads_data == python_data, loads_data

    dumps_data = serializer.dump_pickle(loads_data)
    loads_data = serializer.load_pickle(dumps_data)
    assert dumps_data == pickle_bytes, loads_data
    assert loads_data == python_data, loads_data

    dumps_data = serializer.dump_yaml(python_data)
    loads_data = serializer.load_yaml(dumps_data)
    assert dumps_data == yaml_str, dumps_data
    assert loads_data == python_data

    try:
        serializer = Serializer(path=tmp_file, data_type=dict)
        loads_data = serializer.load_json()
        assert loads_data == dict(), loads_data

        dumps_data = serializer.dump_json(python_data)
        assert dumps_data == json_str, dumps_data
        loads_data = serializer.load_json(trans_key_to_num=True)
        assert loads_data == python_data, loads_data

        dumps_data = serializer.dump_yaml(python_data)
        assert dumps_data == yaml_str, dumps_data
        loads_data = serializer.load_yaml()
        assert loads_data == python_data, loads_data

        dumps_data = serializer.dump_pickle(python_data)
        assert dumps_data == pickle_bytes, dumps_data
        loads_data = serializer.load_pickle()
        assert loads_data == python_data, loads_data
    finally:
        tmp_file.unlink(missing_ok=True)


def tests():
    test_bit_manipulator()
    test_serializer()
    test_compress_serializer()


if __name__ == '__main__':
    tests()



import shutil
import time
import traceback
from pathlib import Path
from typing import Callable, Any


def retry_func(func: Callable, args: tuple | list = None, kwargs: dict = None, max_attempts: int = 3,
               delay: float = 0, retry_exceptions: tuple | list = None, emit_tip: bool = False) -> Any:
    attempts = 0
    args, kwargs = args or (), kwargs or {}
    while max_attempts < 0 or attempts < max_attempts:
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as e:
            if retry_exceptions is None or any(isinstance(e, exc) for exc in retry_exceptions):
                if emit_tip:
                    print(f'{func.__name__} attempts {attempts} times, exception is {type(e).__name__}: {e}')
                attempts += 1
                time.sleep(delay)
            else:
                raise e
    raise Exception(f"Max retry attempts reached for function: {func.__name__}")


class SaveHoi4AutoSave:
    auto_save_file_dir = Path.home() / r'Documents\Paradox Interactive\Hearts of Iron IV\save games'
    auto_save_file_path = auto_save_file_dir / 'autosave.hoi4'

    def __init__(self, start_time: tuple = (36, 2), last_modified_time: float = time.time()):
        self.start_time = list(start_time)
        self.last_modified_time = last_modified_time
        self.monitor_file_modification()

    def monitor_file_modification(self):
        while True:
            modification_time = retry_func(self.auto_save_file_path.stat,
                                           delay=2, max_attempts=-1).st_mtime
            if modification_time > self.last_modified_time:
                self.last_modified_time = modification_time
                self.save()
                self.update_start_time()
            else:
                time.sleep(1)

    def save(self):
        file_name = f'19{self.start_time[0]}-{self.start_time[1]}_{int(time.time())}.hoi4'
        file_path = self.auto_save_file_dir / file_name
        retry_func(shutil.copy2, (self.auto_save_file_path, file_path), delay=1, emit_tip=True)
        print(f'Saved {file_name}')

    def update_start_time(self):
        self.start_time[1] += 1
        if self.start_time[1] == 13:
            self.start_time[1] = 1
            self.start_time[0] += 1


if __name__ == '__main__':
    try:
        s = SaveHoi4AutoSave(start_time=(36, 2))
    except Exception as e:
        traceback.print_exc()
        input()



import functools
import os
import re
from pathlib import Path
from typing import Callable
from typing import List, Union, Generator


def _path_operation(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(path: str | Path) -> str | Path:
        original_type = type(path)
        if isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"Path must be a string or Path object, not {original_type}")

        new_path = func(path)

        if original_type is str:
            return str(new_path)
        return new_path

    return wrapper


@_path_operation
def auto_rename(path: str | Path) -> str | Path:
    """If the file name has existed, add (n) after the file name and return, or return the original name."""
    n = 0
    while path.exists():
        n += 1
        stem = re.sub(r"\(\d+\)$", "", path.stem)
        path = path.with_name(f'{stem}({n}){path.suffix}')
    return path


def get_file_paths(folder_path: str | Path, relative: bool = False, generator: bool = False, to_str: bool = False
                   ) -> Union[List[Path], List[str], Generator[str | Path, None, None]]:
    def _generate_file_paths() -> Generator[str | Path, None, None]:
        for root, dirs, files in os.walk(folder_path):
            for file in files:
                if relative:
                    file_path = os.path.relpath(os.path.join(root, file), folder_path)
                else:
                    file_path = os.path.abspath(os.path.join(root, file))
                if not to_str:
                    file_path = Path(file_path)
                yield file_path

    if generator:
        return _generate_file_paths()
    else:
        return list(_generate_file_paths())



from tests.__init__ import *
from psplpy.middleware_utils import Rabbitmq


Rabbitmq.HOST = get_env('RABBITMQ_HOST')
Rabbitmq.USER = get_env('RABBITMQ_USER')
Rabbitmq.PW = get_env('RABBITMQ_PW')

send_data1 = {'1': 100, '2': False, '3': '你好', '4': 3.14}
send_data2 = {'1': -1, '2': True, '3': 'hello', '4': 2/3}
receiver1_data_list, receiver2_data_list, receiver3_data_list = [], [], []


def callback1(ch, method, properties, body) -> None:
    global receiver1_data_list
    receiver1_data_list.append(body)
    print(body)


def callback2(ch, method, properties, body) -> None:
    global receiver2_data_list
    receiver2_data_list.append(body)
    print(body)


def callback3(ch, method, properties, body) -> None:
    global receiver3_data_list
    receiver3_data_list.append(body)
    print(body)


def tests():
    exchange = 'test'

    rabbitmq_receiver1 = Rabbitmq(serializer=Rabbitmq.JSON, compress=True)
    rabbitmq_receiver1.recv_init(exchange=exchange, binding_keys=['rabbit.*.quickly', 'direct'], callback=callback1)
    rabbitmq_receiver1.start_consuming()

    rabbitmq_receiver1.management.create_vhost('test')
    print(rabbitmq_receiver1.management.list_vhosts())
    assert rabbitmq_receiver1.management.list_vhosts() == ['/', 'test']
    rabbitmq_receiver1.management.delete_vhost('test')
    assert rabbitmq_receiver1.management.list_vhosts() == ['/']

    rabbitmq_receiver2 = Rabbitmq(serializer=Rabbitmq.PICKLE, compress=True)
    # will receive all routing keys' messages like fanout
    rabbitmq_receiver2.recv_init(exchange=exchange, binding_keys=['#'], callback=callback2)
    rabbitmq_receiver2.start_consuming()

    rabbitmq_receiver3 = Rabbitmq(serializer=Rabbitmq.PICKLE, compress=True)
    rabbitmq_receiver3.recv_init(exchange=exchange, binding_keys=['*.jump.*', '#.critical'], callback=callback3)
    rabbitmq_receiver3.start_consuming()

    rabbitmq_sender1 = Rabbitmq(serializer=Rabbitmq.PICKLE, compress=True)
    rabbitmq_sender1.send_init(exchange=exchange, routing_keys=['rabbit.jump.quickly', 'direct'])
    rabbitmq_sender1.basic_publish(send_data1)

    rabbitmq_sender2 = Rabbitmq(serializer=Rabbitmq.PICKLE, compress=True)
    rabbitmq_sender2.send_init(exchange=exchange, routing_keys=['node1.kern.critical', 'direct'])
    rabbitmq_sender2.basic_publish(send_data2)

    # if it fails, increasing the interval
    time.sleep(0.5)
    rabbitmq_receiver1.stop_consuming()
    rabbitmq_receiver1.close(suppress_error=True)
    rabbitmq_receiver2.close(suppress_error=True)
    rabbitmq_receiver3.close(suppress_error=True)

    global receiver1_data_list, receiver2_data_list, receiver3_data_list
    assert (sorted(receiver1_data_list, key=lambda x: x['1']) ==
            sorted([send_data1] * 2 + [send_data2], key=lambda x: x['1'])), receiver1_data_list
    assert (sorted(receiver2_data_list, key=lambda x: x['1']) ==
            sorted([send_data1] * 2 + [send_data2] * 2, key=lambda x: x['1'])), receiver2_data_list
    assert (sorted(receiver3_data_list, key=lambda x: x['1']) ==
            sorted([send_data1] + [send_data2], key=lambda x: x['1'])), receiver3_data_list

    rabbitmq_sender1.channel.exchange_delete(exchange=exchange)
    rabbitmq_sender1.close()
    rabbitmq_sender2.close()


if __name__ == '__main__':
    tests()



from psplpy.file_utils import *


def count_file(file_path: Path, encoding='utf-8', non_empty: bool = False, non_comment: bool = False) -> int:
    lines = file_path.read_text(encoding=encoding).splitlines()
    if non_empty:
        lines = [line for line in lines if line.strip()]
    if non_comment:
        comment_lines = ''
        comment_list = []
        multiline_comment_flag = False
        new_lines = []
        for i in range(len(lines)):
            if i == 0:
                previous_line = ''
            else:
                previous_line = lines[i - 1].strip()
            line = lines[i].strip()
            if multiline_comment_flag:
                # if the multiline comment head had been found in the previous line,
                # next we should find the tail of this comment, once we find it, this multiline comment will over
                comment_lines += lines[i] + '\n'
                if line[-3:] in ['"""', "'''"]:
                    multiline_comment_flag = False
                    comment_list.append((comment_lines[:-1], file_path, i))
                    comment_lines = ''
            else:
                head_length = 0
                if line[:3] in ['"""', "'''"]:
                    head_length = 3
                elif line[:4] in ['r"""', "r'''", 'f"""', "f'''", 'u"""', "u'''"]:
                    head_length = 4
                if head_length:  # if found the multiline comment head
                    has_multiline = False
                    # if equals, that means this line just has the multiline comment head, it must be the beginning
                    if len(line) == head_length:
                        has_multiline = True
                    # if not equals, but the last 3 chars not in them, that means the multiline comment
                    # not just one line, otherwise, the multiline comment only one line
                    elif line[-3:] not in ['"""', "'''"]:
                        has_multiline = True
                    if has_multiline:
                        # if this string turns out to have multiline, but it's previous line's last char in them,
                        # that means this multiline string is a function's parameter, not a comment
                        if (not previous_line) or previous_line[-1] not in ['[', '(', ',']:
                            multiline_comment_flag = True
                            comment_lines += lines[i] + '\n'
                        else:  # because it's a function's parameter, so it's the real code
                            new_lines.append(line)
                    else:  # has not multiline, just a single line comment, continue
                        comment_list.append((lines[i], file_path, i))
                elif line[0] == '#':  # is a single line comment, continue
                    comment_list.append((lines[i], file_path, i))
                elif line[0] in ('"', "'"):
                    if (not previous_line) or previous_line[-1] not in ['[', '(', ',']:
                        comment_list.append((lines[i], file_path, i))
                    else:
                        new_lines.append(line)
                else:
                    new_lines.append(line)
        lines = new_lines
        for comment in comment_list:
            content, path, lineno = comment
            print(f'{path}, {lineno}, {"=" * 50}')
            print(content)
    return len(lines)


def count_dir(project_dir: str | Path, encoding: str = 'utf-8') -> tuple[int, int, int]:
    files = [file for file in get_file_paths(project_dir) if file.suffix == '.py']
    print(f'Total files：{len(files)}')
    count = 0
    for file in files:
        count += count_file(file, encoding)
    print(f'Total lines: {count}')
    non_empty_count = 0
    for file in files:
        non_empty_count += count_file(file, encoding, non_empty=True)
    print(f'After excluded empty lines: {non_empty_count}')
    non_comment_count = 0
    for file in files:
        non_comment_count += count_file(file, encoding, non_empty=True, non_comment=True)
    print(f'After excluded comment lines: {non_comment_count}')
    return count, non_empty_count, non_comment_count


if __name__ == '__main__':
    project_dir = Path(__file__).parent.parent.parent
    print(project_dir)
    count_dir('/usr/local/lib/python3.10/site-packages/django/utils')



import io
import cv2
import numpy as np
from PIL import Image
from tests.__init__ import *
from psplpy.image_utils import Rect, ImgConv, Img, ImgDraw

img_tmp_dir = tmp_dir / 'image'
img_tmp_dir.mkdir(exist_ok=True)
cat_png_path = rc_dir / 'cat.png'


def test_Rect():
    rect = Rect([100.1, 100.9, 200.1, 200.9], rect_format=Rect.LTRB)
    assert Rect([(100.1, 100.9), (200.1, 200.9)], rect_format=Rect.LT_RB) == rect, rect
    assert Rect([100.1, 100.9, 100, 100], rect_format=Rect.LTWH) == rect, rect
    assert Rect([150.1, 150.9, 50.0, 50.0], rect_format=Rect.CENTER_RECT) == rect, rect
    assert Rect([[100.1, 100.9], [200.1, 100.9], [200.1, 200.9], [100.1, 200.9]],
                rect_format=Rect.LT_RT_RB_LB) == rect, rect

    assert rect.to_lt_rb() == [[100.1, 100.9], [200.1, 200.9]], rect.to_lt_rb()
    assert rect.to_ltwh() == [100.1, 100.9, 100, 100], rect.to_ltwh()
    assert rect.to_ltrb() == [100.1, 100.9, 200.1, 200.9], rect.to_ltrb()
    assert rect.to_lt_rt_rb_lt() == [[100.1, 100.9], [200.1, 100.9], [200.1, 200.9], [100.1, 200.9]], rect.to_lt_rt_rb_lt()
    assert rect.to_center() == [150.1, 150.9], rect.to_center()
    assert rect.to_center_rect() == [150.1, 150.9, 50.0, 50.0], rect.to_center_rect()


def test_ImgConv():
    path_img = ImgConv(cat_png_path)
    img_list = [path_img, ImgConv(Image.open(cat_png_path)), ImgConv(path_img.to_numpy()),
                ImgConv(cv2.imread(str(cat_png_path)), from_opencv=True),
                ImgConv(path_img.to_bytesio()), ImgConv(path_img.to_bytes()), ImgConv(path_img.to_bytearray())]

    def _test(img: Any, i: int) -> None:
        assert isinstance(img.to_pil(), Image.Image), i
        assert isinstance(img.to_numpy(), np.ndarray), i
        assert isinstance(img.to_opencv(), np.ndarray), i
        assert isinstance(img.to_bytesio(), io.BytesIO), i
        assert isinstance(img.to_bytes(), bytes), i
        assert isinstance(img.to_bytearray(), bytearray), i
        path = img_tmp_dir / f'cat{i}.png'
        assert img.to_path(path) == path, i

    for i, img in enumerate(img_list):
        multiprocessing.Process(target=_test, args=(img, i)).start()


def test_Img():
    img = Img(cat_png_path)
    img.resize(ratio=0.5).to_path(img_tmp_dir / 'cat_resize.png')
    img = Img(cat_png_path)
    img.contrast(2).to_path(img_tmp_dir / 'cat_contrast.png')
    img = Img(cat_png_path)
    img.brightness(2).to_path(img_tmp_dir / 'cat_brightness.png')
    img = Img(cat_png_path)
    img.sharpness(2).to_path(img_tmp_dir / 'cat_sharp.png')
    img = Img(cat_png_path)
    img.grayscale().to_path(img_tmp_dir / 'cat_grayscale.png')
    img = Img(cat_png_path)
    img.invert().to_path(img_tmp_dir / 'cat_invert.png')
    img = Img(cat_png_path)
    img.rotate(120).to_path(img_tmp_dir / 'cat_rotate.png')
    img = Img(cat_png_path)
    img.binaryzation(128).to_path(img_tmp_dir / 'cat_binaryzation.png')
    img = Img(cat_png_path)
    img.crop(Rect([0, 0, *img.image.size]).resize(0.5)).to_path(img_tmp_dir / 'cat_crop.png')
    img = Img(cat_png_path)
    img.dpi(1000).to_path(img_tmp_dir / 'cat_dpi.png')
    img = Img(cat_png_path)
    img.resize(ratio=0.5).contrast(2).brightness(2).sharpness(2).to_path(img_tmp_dir / 'cat_combination.png')


def test_ImgDraw():
    img_draw = ImgDraw(cat_png_path)
    rects = [Rect([0, 0, *img_draw.image.size]).resize(0.5),
             Rect([300, 300, 50, 50])]
    img_draw.rectangles(rects).to_path(img_tmp_dir / 'cat_rectangle.png')

    img_draw = ImgDraw(cat_png_path)
    polygons = [[(50, 50), (75, 75), (60, 90)],
                [(100, 100), (150, 105), (130, 150), (145, 200), (95, 250)]]
    img_draw.polygons(polygons).to_path(img_tmp_dir / 'cat_polygon.png')

    img_draw = ImgDraw(cat_png_path)
    text = 'Hello,\n 我來自北方中國大陆，\n一個城市名字是北京，\n作为中國的首都'
    img_draw.text(text, left_top_point=(100, 200)).to_path(img_tmp_dir / 'cat_text.png')

    img_draw = ImgDraw(cat_png_path)
    text = 'Hello, 我來自北方中國大陆，\n一個城市名字是北京，作为中國的首都'
    rect = Rect([100, 100, 1000, 250], rect_format=Rect.LTRB)
    img_draw.rectangles([rect])
    img_draw.text_in_rect(text, rect, fill='yellow').to_path(img_tmp_dir / 'cat_text_in_rect.png')


def tests():
    print('test image utils')
    test_Rect()
    test_ImgConv()
    test_Img()
    test_ImgDraw()


if __name__ == '__main__':
    tests()



"""
Above Python3.9, the labelImg will collapse
"""

import site
from pathlib import Path

for path in site.getsitepackages():
    canvas_path = Path(path) / 'libs/canvas.py'
    print(canvas_path)
    if canvas_path.exists():
        content = canvas_path.read_text(encoding='utf-8')
        content = content.replace('p.drawRect(left_top.x(), left_top.y(), rect_width, rect_height)',
                                  'p.drawRect(int(left_top.x()), int(left_top.y()), int(rect_width), int(rect_height))')
        content = content.replace('p.drawLine(self.prev_point.x(), 0, self.prev_point.x(), self.pixmap.height())',
                                  'p.drawLine(int(self.prev_point.x()), 0, int(self.prev_point.x()), int(self.pixmap.height()))')
        content = content.replace('p.drawLine(0, self.prev_point.y(), self.pixmap.width(), self.prev_point.y())',
                                  'p.drawLine(0, int(self.prev_point.y()), int(self.pixmap.width()), int(self.prev_point.y()))')
        canvas_path.write_text(content, encoding='utf-8')
        print(f'Fix over')
        break
else:
    raise FileNotFoundError(f"Canvas file not found in site-packages")



from concurrent import futures

from tests.__init__ import *
from psplpy.network_utils import *


def get_ip_address():
    try:
        temp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        temp_socket.connect(("114.114.114.114", 80))  # connect to a public ip address
        ip_address = temp_socket.getsockname()[0]  # get the host's ip
        temp_socket.close()
        return ip_address
    except socket.error:
        return None


port = 12345
host_ip = get_ip_address()
client_port = find_free_port(host_ip, try_range=(12345, 12999))
data = b"Hello World" * 1024 * 32


def sender():
    def handler(client_socket: ClientSocket, addr):
        print(f'client {addr}')
        assert addr == (host_ip, client_port)

        received_data = client_socket.recv()
        assert received_data == data
        client_socket.send(data)
        recv_tmp_file = tmp_dir / 'recv_tmp.tmp'
        client_socket.recvf(recv_tmp_file)
        assert tmp_file.read_text() == recv_tmp_file.read_text()

        tmp_file.unlink()
        recv_tmp_file.unlink()
        client_socket.close()
        server.close()

    server = ServerSocket(port=port)
    server.handle(handler)


def recver():
    client = ClientSocket(port=port, client_host=host_ip, client_port=client_port)
    client.connect()
    client.send(data)
    received_data = client.recv()
    assert received_data == data
    tmp_file.write_bytes(data)
    client.sendf(tmp_file)
    client.close()


def test_find():
    def handler(client_socket: ClientSocket, addr):
        client_socket.close()

    test_port = find_free_port(host_ip, [12345, ], [12346, ], (12345, 12999))
    print(f'test port {test_port}')
    s = ServerSocket(host=host_ip, port=test_port)
    s.handle(handler)
    assert find_running_port(host=host_ip, try_range=(test_port - 100, test_port + 100)) == test_port
    assert find_running_port(host=host_ip, try_ports=[test_port, ], exclude_ports=[test_port, ],
                             try_range=(test_port - 100, test_port + 100)) is None

    s.close()


def test_mp_http():
    class TestServer(MpHttpServer):
        def init(self) -> None:
            self.num = 1

        def main_loop(self, data: Any) -> Any:
            time.sleep(data / 1)
            return {'result': data, 'num': self.num}

    def show_load(c: MpHttpClient):
        time.sleep(0.25)
        return c.get_load()

    s = TestServer(port=find_free_port(), workers=6, max_fetch_timeout=2, result_timeout=1).run_server(new_thread=True)
    time.sleep(0.25)
    c = MpHttpClient(server=s)

    # test get_load and batch
    load_results = futures.ThreadPoolExecutor().map(show_load, [c])  # starting a new thread to get the load
    result = c.batch([0.5, 0.75, 1])
    load_results = list(load_results)
    assert load_results == [0.5], load_results
    assert result == [{'result': 0.5, 'num': 1}, {'result': 0.75, 'num': 1}, {'result': 1, 'num': 1}], result

    # test get_progress
    data_list = [0.5, 0.75, 1]
    task_id = c.submit(data_list)
    assert isinstance(task_id, int)
    task_progress = c.get_progress(task_id)
    assert task_progress == (0, len(data_list)), task_progress
    time.sleep(0.75 + 0.05)
    task_progress = c.get_progress(task_id)
    assert task_progress == (2, len(data_list)), task_progress
    time.sleep(1 - 0.75)

    # test task result expiration and task fetch timeout
    time.sleep(2)
    try:
        result = c.fetch(task_id)
    except KeyError as e:
        print(e)
    else:
        assert False
    task_id = c.submit([3])
    try:
        result = c.fetch(task_id)
    except TimeoutError as e:
        print(e)
    else:
        assert False
    time.sleep(1.5)

    # test no load
    assert c.get_load() == 0
    time.sleep(0.25)

    # test send empty list
    task_id = c.submit([])
    assert c.fetch(task_id) == []
    s.close_server()


def tests():
    with futures.ThreadPoolExecutor() as executor:
        sender_future = executor.submit(sender)
        time.sleep(0.1)
        recver_future = executor.submit(recver)
        time.sleep(0.1)
        sender_future.result(), recver_future.result()
    test_find()
    test_mp_http()


if __name__ == '__main__':
    tests()



import time
from typing import Any
from psplpy.network_utils import ServerSocket, ClientSocket
from psplpy.other_utils import overlay_print


class SpeedTestServer:
    def __init__(self, server_socket: ServerSocket):
        self.server_socket = server_socket

    @staticmethod
    def handler(client_socket: ClientSocket, addr: Any):
        print(f'Connection from {addr}')
        t_start = time.time()
        total_length = 0
        while True:
            this_t_start = time.time()
            data = client_socket.recv()
            if not data:
                break
            total_length += len(data)
            total_mb = total_length / (1024 * 1024)
            speed = total_mb / (time.time() - t_start)
            current_speed = (len(data) / (1024 * 1024)) / (time.time() - this_t_start)
            overlay_print(f'Received: {total_mb:.2f} mb, avg speed {speed:.2f} mb/s, '
                          f'current speed {current_speed:.2f} mb/s')
        print(f'\nConnection from {addr} closed')

    def start(self):
        print(f'Starting server on ({self.server_socket.host}, {self.server_socket.port})')
        self.server_socket.handle(self.handler)


class SpeedTestClient:
    def __init__(self, client_socket: ClientSocket, data_size_mb: float = 1):
        self.client_socket = client_socket
        self.test_data = b'10' * int(512 * 1024 * data_size_mb)

    def start(self, test_seconds: float = 30):
        self.client_socket.connect()
        total_length = 0
        t_start = time.time()
        while time.time() - t_start < test_seconds:
            self.client_socket.send(self.test_data)
            total_length += len(self.test_data)
            total_mb = total_length / (1024 * 1024)
            overlay_print(f'Sent {total_mb} mb')
        print('Test finished')


if __name__ == '__main__':
    SpeedTestServer(ServerSocket()).start()



import re
from pathlib import Path
from typing import Any, Callable
import cv2
import numpy as np
from psplpy.image_utils import ImgConv, Rect, ImgDraw


class ImgDet:
    TM_CCOEFF_NORMED = cv2.TM_CCOEFF_NORMED
    TM_CCORR_NORMED = cv2.TM_CCORR_NORMED
    TM_SQDIFF_NORMED = cv2.TM_SQDIFF_NORMED

    @staticmethod
    def _filter_rects(rects: list[Rect], threshold: float) -> list[Rect]:
        """if the two rectangles overlap percentage greater than the threshold, remove the second rectangle"""
        if 0 <= threshold <= 1:
            marked_rects = [[rect, True] for rect in rects]

            for i in range(len(marked_rects)):
                if marked_rects[i][1]:
                    for j in range(i + 1, len(marked_rects)):
                        if marked_rects[j][1]:
                            overlap_ratio = max(marked_rects[i][0].overlap_pct(marked_rects[j][0]))
                            if overlap_ratio >= threshold:
                                marked_rects[j][1] = False

            return [rect[0] for rect in marked_rects if rect[1]]
        return rects

    @staticmethod
    def _get_result(image: Any, template_image: Any, method: int = cv2.TM_CCOEFF_NORMED, save_match: Path | str = None):
        cvt_image = ImgConv(image).to_opencv()
        cvt_template_image = ImgConv(template_image).to_opencv()
        result = cv2.matchTemplate(cvt_image, cvt_template_image, method)  # 使用模板匹配
        if method == ImgDet.TM_SQDIFF_NORMED:
            result = -((result - 0.5) * 2)
        if save_match is not None:
            result_copy = result.copy()
            result_copy = ((result_copy + 1) * (255 / 2)).astype(np.uint8)
            cv2.imwrite(str(save_match), result_copy)
        return result, cvt_image, cvt_template_image

    @staticmethod
    def _draw_result(cvt_image: np.ndarray, rects: list[Rect], path: Path | str) -> None:
        if path is not None:
            copied_image = cvt_image.copy()
            for rect in rects:
                cv2.rectangle(copied_image, rect[0], rect[1], (0, 0, 255), 1)
            cv2.imwrite(str(path), copied_image)

    @staticmethod
    def get_box(image: Any, template_image: Any, confidence: float = 0.9, method: int = cv2.TM_CCOEFF_NORMED,
                draw_result: Path | str = None, save_match: Path | str = None) -> tuple[Rect, float] | None:
        """return: (Rect, confidence) or None"""
        result, cvt_image, cvt_template_image = ImgDet._get_result(image, template_image, method, save_match)

        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
        if max_val < confidence:
            return None
        top_left = max_loc
        h, w = cvt_template_image.shape[:2]
        rect = Rect((top_left, (top_left[0] + w, top_left[1] + h)), rect_format=Rect.LT_RB)
        ImgDet._draw_result(cvt_image, [rect], draw_result)
        return rect, float(result[top_left[::-1]])

    @staticmethod
    def get_boxes(image: Any, template_image: Any, confidence: float = 0.9, overlap: float = 0.5,
                  method: int = cv2.TM_CCOEFF_NORMED, draw_result: Path | str = None,
                  save_match: Path | str = None) -> list[tuple[Rect, float]]:
        """return: [(Rect1, confidence1), (Rect2, confidence2), ...] or []"""
        result, cvt_image, cvt_template_image = ImgDet._get_result(image, template_image, method, save_match)
        indices = np.where(result >= confidence)
        positions = np.column_stack(indices)
        sorted_positions = np.argsort(result[indices])[::-1]  # sort in descending order
        sorted_positions = [sorted_position[::-1] for sorted_position in positions[sorted_positions]]

        h, w = cvt_template_image.shape[:2]
        boxes = [Rect([top_left, (top_left[0] + w, top_left[1] + h)], rect_format=Rect.LT_RB)
                 for top_left in sorted_positions]
        boxes = ImgDet._filter_rects(boxes, overlap)
        rects = [(box, float(result[tuple(box[0][::-1])])) for box in boxes]
        ImgDet._draw_result(cvt_image, [rect[0] for rect in rects], draw_result)
        return rects

    @staticmethod
    def det_img(image: Any, template_image: Any, confidence: float = 0.9, method: int = cv2.TM_CCOEFF_NORMED,
                draw_result: Path | str = None, save_match: Path | str = None) -> bool:
        return bool(ImgDet.get_box(image, template_image, confidence, method, draw_result, save_match))


class TxtProcessor:
    NUM = 'num'
    EN = 'en'
    NO_SPACE = 'no_space'
    CASEFOLD = 'casefold'

    def __init__(self, *flags: str | Callable):
        self.flags = flags

    def process(self, text: str) -> str:
        for flag in self.flags:
            if isinstance(flag, str):
                text = getattr(self, flag)(text)
            elif callable(flag):
                text = flag(text)
            else:
                raise AssertionError
        return text

    @staticmethod
    def casefold(text: str) -> str:
        return text.casefold()

    @staticmethod
    def num(text: str) -> str:
        return re.sub(r'\D', '', text)

    @staticmethod
    def en(text: str) -> str:
        return re.sub(r'[^a-zA-Z]', '', text)

    @staticmethod
    def no_space(text: str) -> str:
        return re.sub(r'\s+', '', text)

    @staticmethod
    def camel_to_snake(camel_case: str) -> str:
        # 将驼峰字符串转换为下划线风格
        snake_case = re.sub(r'(?<!^)(?=[A-Z])', '_', camel_case).lower()
        return snake_case


class TextInfo:
    def __init__(self, region: list, text: str, confidence: float):
        self.rect = ImgDraw.bounding_rect(region)
        self.text = text
        self.confidence = confidence

    def __iter__(self):
        for item in [self.rect, self.text, self.confidence]:
            yield item

    @staticmethod
    def _index_mapping(index) -> str:
        index_mapping = {0: 'rect', 1: 'text', 2: 'confidence',
                         -3: 'rect', -2: 'text', -1: 'confidence'}
        if index in index_mapping:
            return index_mapping[index]
        raise IndexError('Index out of range.')

    def __getitem__(self, index):
        if isinstance(index, slice):  # if slice
            return [self[i] for i in range(*index.indices(3))]
        return getattr(self, self._index_mapping(index))

    def __setitem__(self, index, value):
        setattr(self, self._index_mapping(index), value)

    def __str__(self):
        return f'({self.rect}, {self.text}, {self.confidence:.4f})'

    __repr__ = __str__

    def __eq__(self, other):
        if isinstance(other, TextInfo):
            return list(self) == list(other)
        return False


class TextInfoList:
    def __init__(self, text_info_list: list):
        self.text_info_list = self._get_text_info_list(text_info_list)

    @staticmethod
    def _get_text_info_list(text_info_list: list) -> list[TextInfo]:
        self_text_info_list = []
        for info in text_info_list:
            if isinstance(info, TextInfo):
                self_text_info_list.append(info)
            else:
                self_text_info_list.append(TextInfo(*info))
        return self_text_info_list

    def __iter__(self):
        for text_info in self.text_info_list:
            yield text_info

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.text_info_list[index.start:index.stop:index.step]
        return self.text_info_list[index]

    def __setitem__(self, index, value):
        self.text_info_list[index] = value

    def __str__(self):
        return str(self.text_info_list)

    __repr__ = __str__

    def __eq__(self, other):
        if isinstance(other, TextInfoList):
            return self.text_info_list == other.text_info_list
        return False

    def text_list(self) -> list[str]:
        return [info.text for info in self.text_info_list]

    def merged_text(self) -> str:
        return ''.join(self.text_list())


class TxtDet:
    IN = 'in'
    IS = 'is'

    def __init__(self, txt_processor: TxtProcessor = None, process_text: bool = True):
        self.txt_processor = txt_processor or TxtProcessor()
        self.process_text = process_text    # whether process the 'text' argument

    @staticmethod
    def get_txt(text_info_list: list | TextInfoList, rect: Any, overlap: float = 0.75) -> TextInfo:
        rect = Rect(rect)
        for text_info in TextInfoList(text_info_list):
            percentage = min(rect.overlap_pct(text_info.rect))
            if percentage >= overlap:
                return text_info

    def _get_condition(self, text: str | re.Pattern, mode: str) -> Callable:
        if self.process_text:
            text = self.txt_processor.process(text)
        if isinstance(text, str):
            if mode == self.IN:
                return lambda text_info: text in text_info.text
            elif mode == self.IS:
                return lambda text_info: text == text_info.text
        elif isinstance(text, re.Pattern):
            return lambda text_info: text.search(text_info.text)
        raise AssertionError

    def _txt_process(self, text_info_list: list | TextInfoList) -> TextInfoList:
        if not isinstance(text_info_list, TextInfoList):
            text_info_list = TextInfoList(text_info_list)
        for info in text_info_list:
            info.text = self.txt_processor.process(info.text)
        return text_info_list

    def _get_boxes(self, text_info_list: list | TextInfoList, text: str | re.Pattern, get_one: bool = False,
                   mode: str = IN) -> TextInfoList | TextInfo:
        condition = self._get_condition(text, mode)
        result_list = TextInfoList([])
        print(self._txt_process(text_info_list))
        for original_info, processed_info in zip(TextInfoList(text_info_list), self._txt_process(text_info_list)):
            if condition(processed_info):
                if get_one:
                    return original_info
                result_list.text_info_list.append(original_info)
        return result_list

    def get_boxes(self, text_info_list: list | TextInfoList, text: str | re.Pattern,
                  mode: str = IN) -> TextInfoList:
        return self._get_boxes(text_info_list, text, mode=mode)

    def get_box(self, text_info_list: list | TextInfoList, text: str | re.Pattern, mode: str = IN) -> TextInfo | None:
        return self._get_boxes(text_info_list, text, get_one=True, mode=mode)

    class MatchObj:
        def __init__(self, text_info_list: list | TextInfoList, matched_text_info_list: list | TextInfoList):
            self.text_info_list = text_info_list
            self.matched_text_info_list = matched_text_info_list
            self.result_list = []

        @property
        def match_list(self) -> list[TextInfo]:
            return [self.matched_text_info_list[i] for i in range(len(self.result_list)) if self.result_list[i]]

        @property
        def mismatch_list(self) -> list[TextInfo]:
            return [self.matched_text_info_list[i] for i in range(len(self.result_list)) if not self.result_list[i]]

        def _update_bool_and_confidence(self, expected_confidence: float) -> None:
            self.confidence = self.result_list.count(True) / len(self.result_list)
            self.bool = self.confidence >= expected_confidence

    def match(self, text_info_list: list | TextInfoList, matched_text_info_list: list | TextInfoList,
              overlap: float = 0.75, confidence: float = 1, mode: str = IN) -> MatchObj:
        match_obj = self.MatchObj(text_info_list, matched_text_info_list)
        text_info_list, processed_text_info_list = TextInfoList(text_info_list), self._txt_process(text_info_list)

        for matched_text_info in TextInfoList(matched_text_info_list):
            flag = False
            condition = self._get_condition(matched_text_info.text, mode)
            for original_info, processed_info in zip(text_info_list, processed_text_info_list):
                if condition(processed_info):
                    if min(matched_text_info.rect.overlap_pct(original_info.rect)) >= overlap:
                        flag = True
                        break
            match_obj.result_list.append(flag)
        match_obj._update_bool_and_confidence(confidence)
        return match_obj



import functools
import io
import json
import lzma
import math
import pickle
import yaml
import re
import zlib
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Literal, Union
import operator
from psplpy.other_utils import get_key_from_value


class _ComparisonMeta(type):
    def __new__(cls, name, bases, dct):
        # Using this way can avoid the function in loop using the last iteration's value
        # or using a default kwarg, but that would create an unnecessary kwarg
        def create_func(op):
            def method(self, other):
                if isinstance(other, self.__class__):
                    return getattr(operator, op)(self.bit_value, other.bit_value)
                elif isinstance(other, int):
                    return getattr(operator, op)(self.bit_value, other)
                elif isinstance(other, bytes):
                    return getattr(operator, op)(self.bit_value, int.from_bytes(other, byteorder='big'))
                return NotImplemented   # __eq__ and __ne__ will return True or False, can't reach there
            return method

        for op in ['lt', 'le', 'gt', 'ge', 'eq', 'ne']:
            dct[f'__{op}__'] = create_func(op)
        return super().__new__(cls, name, bases, dct)


class BitManipulator(metaclass=_ComparisonMeta):
    _EXCEED_BITS_ERR_INFO = "Bit value's length must be less than or equal to {}"

    def __init__(self, bit_value: Union[int, bytes, 'BitManipulator'] = 0,
                 max_bits: int = None, fixed_bits: int = None):
        self.max_bits = max_bits
        self.fixed_bits = fixed_bits
        if isinstance(bit_value, bytes):
            bit_value = int.from_bytes(bit_value, byteorder='big')
        if isinstance(bit_value, BitManipulator):
            bit_value = bit_value.bit_value
        if isinstance(bit_value, int):
            if bit_value < 0:
                raise ValueError("Bit value must be a non-negative integer")
            self._set_bit_value(bit_value)
        else:
            raise ValueError("Bit value must be an int or bytes object")

    def _check_bits(self, length: int) -> None:
        if self.max_bits is not None and length > self.max_bits:
            raise ValueError(self._EXCEED_BITS_ERR_INFO.format('max_bits'))
        if self.fixed_bits is not None and length > self.fixed_bits:
            raise ValueError(self._EXCEED_BITS_ERR_INFO.format('fixed_bits'))

    def _set_bit_value(self, bit_value: int) -> None:
        self._check_bits(bit_value.bit_length())
        self.bit_value = bit_value

    def __len__(self):
        return self.bit_value.bit_length()

    def _index_check(self, index: int) -> int:
        if index < 0:
            index += len(self)
            if index < 0:
                raise IndexError("Index out of range")
        return index

    def _get_iter_range(self, key: slice) -> list:
        return list(range(*key.indices(key.stop if key.stop > 0 else len(self))))

    def __getitem__(self, key: int | slice) -> str:
        if isinstance(key, slice):
            return ''.join([self[i] for i in self._get_iter_range(key)])
        elif isinstance(key, int):
            return str((self.bit_value >> self._index_check(key)) & 1)
        raise TypeError("Invalid argument type")

    def __setitem__(self, key: int | slice, value: int | str) -> None:
        def _set_bit(index: int, bit: int) -> None:
            if bit == 1:
                self._set_bit_value(self.bit_value | (1 << index))  # Set bit to 1
            elif bit == 0:
                self._set_bit_value(self.bit_value & ~(1 << index))  # Set bit to 0
            else:
                raise ValueError("Value must be 0 or 1")

        if isinstance(key, slice):
            iter_range = self._get_iter_range(key)
            if isinstance(value, int):
                value = bin(value)[2:]
            if len(value) != len(iter_range):
                raise ValueError(f"Slice length {len(iter_range)} not equal to the value length {len(value)}")
            original_bit_value = self.bit_value
            try:
                for i, bit_index in enumerate(iter_range):
                    _set_bit(bit_index, int(value[i]))
            except ValueError as e:
                self._set_bit_value(original_bit_value)
                raise e
        elif isinstance(key, int):
            _set_bit(self._index_check(key), int(value))
        else:
            raise TypeError("Invalid argument type")

    def get_bit(self, key: int | slice) -> str:
        return self[key]

    def set_bit(self, key: int | slice, value: int | str | bytes) -> None:
        self[key] = value

    def to_bytes(self, length: int = None) -> bytes:
        """Convert the bit value to bytes. The length specifies the number of bytes."""
        if length is None:
            bit_length = len(self)
            if self.fixed_bits is not None:
                bit_length = self.fixed_bits
            length = math.ceil(bit_length / 8)
        return self.bit_value.to_bytes(length, byteorder='big')

    def __format__(self, format_spec: Literal["b", "o", "x", "d"]):
        """Support custom formatting: 'b' for binary, 'o' for octal, 'x' for hex, 'd' for decimal."""
        if format_spec in ["b", "o", "x", "d"]:
            if self.fixed_bits and format_spec == "b":
                return f'{self.bit_value:{format_spec}}'.zfill(self.fixed_bits)
            return f'{self.bit_value:{format_spec}}'
        return str(self)

    def __str__(self):
        # Return the bit value as a binary string, MSB first
        return self.__format__('b')

    __repr__ = __str__

    def __iter__(self):
        for position in range(len(self)):
            yield self[position]


class Compressor:
    LZMA = 'lzma'
    ZLIB = 'zlib'

    def __init__(self, lib: str = ZLIB):
        self.lib = lib

    def compress(self, data: bytes, lib: str = '') -> bytes:
        lib = lib or self.lib
        compressed_data = globals()[lib].compress(data)
        return compressed_data

    def decompress(self, compressed_data: bytes, lib: str = '') -> bytes:
        lib = lib or self.lib
        data = globals()[lib].decompress(compressed_data)
        return data


def _check_int_or_float(input_str: str) -> type:
    int_pattern = r'^[-+]?\d+$'
    float_pattern = r'^[-+]?\d+(\.\d+)?$'

    if re.match(int_pattern, input_str):
        return int
    elif re.match(float_pattern, input_str):
        return float
    else:
        return str


def _convert_json_dict_key_to_number(data: Any) -> Any:
    if isinstance(data, dict):
        # if data type is dict, convert it
        converted_dict = {}
        for key, value in data.items():
            if type(key) == str:
                trans_type = _check_int_or_float(key)
                key = trans_type(key)
            # process the values in dict, using recursion
            value = _convert_json_dict_key_to_number(value)
            converted_dict[key] = value
        return converted_dict
    elif isinstance(data, (list, tuple, set)):
        # if date type is list, tuple or set, process it recursively
        converted_list = []
        for item in data:
            converted_item = _convert_json_dict_key_to_number(item)
            converted_list.append(converted_item)
        return type(data)(converted_list)
    else:
        # if it's other type, don't process
        return data


def _get_empty_data_structure(data_type: type | None) -> dict | list | tuple | set | None:
    if data_type is None:
        return None
    types = (dict, list, tuple, set)
    if data_type in types:
        return data_type()
    else:
        raise TypeError(f"Unsupported data type {data_type}")


def _get_data(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(self, data: str | bytes = None, *args, **kwargs) -> Any:
        if not data:
            if not self.path:
                raise AssertionError('For loading data, please provide the data or file path.')
            try:
                data = Path(self.path).read_bytes()
            except FileNotFoundError:  # when file not found
                return _get_empty_data_structure(self.data_type)
        return func(self, data, *args, **kwargs)
    return wrapper


def _dump(self: Any, data: bytes | str) -> bytes | str:
    if self.path:
        if isinstance(data, str):
            Path(self.path).write_text(data, encoding=self.encoding)
        else:
            Path(self.path).write_bytes(data)
    return data


class Serializer:
    def __init__(self, path: str | Path = None, encoding: str = 'utf-8', data_type: type = None):
        self.path, self.encoding, self.data_type = path, encoding, data_type

    @_get_data
    def _load(self, data: str | bytes, lib: ModuleType) -> Any:
        if lib in [json, yaml] and isinstance(data, bytes):
            data = data.decode(self.encoding)
        if lib is json:
            try:
                deserialized_data = json.loads(data)
            except json.decoder.JSONDecodeError:  # when data is empty
                return _get_empty_data_structure(self.data_type)
        elif lib is yaml:
            deserialized_data = yaml.safe_load(data)
        elif lib is pickle:
            try:
                deserialized_data = pickle.loads(data)
            except EOFError:  # when data is empty
                return _get_empty_data_structure(self.data_type)
        else:
            raise AssertionError
        return deserialized_data

    def load_yaml(self, data: str = None) -> Any:
        return self._load(data, yaml)

    def load_json(self, data: str = None, trans_key_to_num: bool = False) -> Any:
        json_data = self._load(data, json)
        if trans_key_to_num:
            return _convert_json_dict_key_to_number(json_data)
        return json_data

    def load_pickle(self, data: bytes = None) -> Any:
        return self._load(data, pickle)

    def dump_yaml(self, data: Any, allow_unicode: bool = True) -> str:
        string_io = io.StringIO()
        yaml.dump(data, string_io, allow_unicode=allow_unicode)
        data = string_io.getvalue()
        return _dump(self, data)

    def dump_json(self, data: Any, indent: int = 4, ensure_ascii: bool = False, minimum: bool = True) -> str:
        kwargs = {'ensure_ascii': ensure_ascii}
        if minimum:
            kwargs['separators'] = (',', ':')
        else:
            kwargs['indent'] = indent
        data = json.dumps(data, **kwargs)
        return _dump(self, data)

    def dump_pickle(self, data: Any) -> bytes:
        data = pickle.dumps(data)
        return _dump(self, data)


class CompressSerializer:
    _UNCOMPRESSED, _COMPRESSED = '0', '1'
    _ZLIB, _LZMA = '0', '1'
    _JSON, _PICKLE, _YAML = '00', '01', '10'
    _LIB_DICT = {_ZLIB: Compressor.ZLIB, _LZMA: Compressor.LZMA}
    _FORMAT_DICT = {_JSON: 'json', _PICKLE: 'pickle', _YAML: 'yaml'}
    AUTO = 'auto'

    class _Metadata:
        _ACCESS_DICT = {'compressed': (0, 1), 'lib': (1, 2), 'format': (2, 4)}

        def __init__(self, metadata: int | bytes = 0):
            self.bm = BitManipulator(metadata, fixed_bits=8)

        def _get(self, start: int, stop: int) -> str:
            return self.bm[start:stop]

        def _set(self, start: int, stop: int, value: str) -> None:
            self.bm[start:stop] = value

        def __getattr__(self, name):
            if name in self._ACCESS_DICT:
                return self._get(*self._ACCESS_DICT[name])
            return super().__getattribute__(name)

        def __setattr__(self, name, value):
            if name in self._ACCESS_DICT:
                return self._set(*self._ACCESS_DICT[name], value)
            super().__setattr__(name, value)

    def __init__(self, path: str | Path = None, encoding: str = 'utf-8', data_type: type = None,
                 compress: bool | str = AUTO, threshold: int = 1024 * 128, compress_lib: str = Compressor.ZLIB):
        """When the data length is greater than the threshold, will execute compression"""
        self.path, self.encoding, self.data_type = path, encoding, data_type
        self.compress, self.threshold, self.compress_lib = compress, threshold, compress_lib
        self._c = Compressor(lib=self.compress_lib)
        self._s = Serializer(encoding=self.encoding, data_type=self.data_type)

    def _get_uncompressed_data_and_format(self, data: bytes) -> tuple[bytes, str]:
        metadata = self._Metadata(data[0])
        data = data[1:]
        if metadata.compressed == self._COMPRESSED:
            data = self._c.decompress(data, lib=self._LIB_DICT[metadata.lib])
        return data, metadata.format

    @_get_data
    def load(self, data: bytes = None, *args, **kwargs) -> Any:
        uncompressed_data, format = self._get_uncompressed_data_and_format(data)
        return getattr(self._s, f'load_{self._FORMAT_DICT[format]}')(uncompressed_data, *args, **kwargs)

    @_get_data
    def load_pickle(self, data: bytes = None) -> Any:
        return self._s.load_pickle(self._get_uncompressed_data_and_format(data)[0])

    @_get_data
    def load_json(self, data: bytes = None, trans_key_to_num: bool = False) -> Any:
        return self._s.load_json(self._get_uncompressed_data_and_format(data)[0].decode(encoding=self.encoding),
                                 trans_key_to_num)

    @_get_data
    def load_yaml(self, data: bytes = None) -> Any:
        return self._s.load_yaml(self._get_uncompressed_data_and_format(data)[0].decode(encoding=self.encoding))

    def _compress_or_not(self, data: bytes, compress: bool | None) -> bool:
        if compress is None:  # if None, depends on self.compress
            if self.compress == self.AUTO and len(data) > self.threshold:
                return True
            return bool(self.compress)
        return bool(compress)

    def _get_compressed_data(self, data: bytes, compress: bool | None, format: str) -> bytes:
        metadata = self._Metadata()
        metadata.format = format
        metadata.lib = get_key_from_value(self._LIB_DICT, self.compress_lib)

        compress = self._compress_or_not(data, compress)
        if compress:
            metadata.compressed = self._COMPRESSED
            data = self._c.compress(data)
        else:
            metadata.compressed = self._UNCOMPRESSED
        return _dump(self, metadata.bm.to_bytes() + data)

    def dump_pickle(self, data: Any, compress: bool | None = None) -> bytes:
        return self._get_compressed_data(self._s.dump_pickle(data), compress, self._PICKLE)

    def dump_json(self, data: Any, compress: bool | None = None,
                  indent: int = 4, ensure_ascii: bool = False, minimum: bool = True) -> bytes:
        data = self._s.dump_json(data, indent, ensure_ascii, minimum)
        return self._get_compressed_data(data.encode(encoding=self.encoding), compress, self._JSON)

    def dump_yaml(self, data: Any, compress: bool | None = None, allow_unicode: bool = True) -> bytes:
        data = self._s.dump_yaml(data, allow_unicode)
        return self._get_compressed_data(data.encode(encoding=self.encoding), compress, self._YAML)



import argparse
import threading
import time


class ThreadLocks:
    def __init__(self):
        self.locks = argparse.Namespace()
        # lock_name: {'auto_release': bool, 'timeout': float, 'last_acquired_time': time.time}
        self.auto_release_info = {}

    def has_lock(self, lock_name: str) -> bool:
        return hasattr(self.locks, lock_name)

    def _get_lock(self, lock_name: str) -> threading.Lock:
        return getattr(self.locks, lock_name)

    def set_lock(self, lock_name: str, auto_release_time: float = -1) -> None:
        if self.has_lock(lock_name):
            raise AttributeError(f"{lock_name} has already existed")
        else:
            setattr(self.locks, lock_name, threading.Lock())
        if auto_release_time >= 0:
            info = {'auto_release': True, 'timeout': auto_release_time, 'last_acquired_time': 0}
            self.auto_release_info[lock_name] = info
            self._auto_release_daemon(lock_name)
        else:
            self.auto_release_info[lock_name] = {'auto_release': False}

    def acquire_lock(self, lock_name: str, blocking: bool = True, timeout: float = -1) -> bool:
        lock = self._get_lock(lock_name)
        acquire_result = lock.acquire(blocking=blocking, timeout=timeout)
        if acquire_result:
            self.auto_release_info[lock_name]['last_acquired_time'] = time.time()
        return acquire_result

    def release_lock(self, lock_name: str, ignore_released: bool = True) -> None:
        lock = self._get_lock(lock_name)
        if ignore_released and not self.locked(lock_name):
            return None
        return lock.release()

    def locked(self, lock_name: str) -> bool:
        lock = self._get_lock(lock_name)
        return lock.locked()

    def _auto_release_daemon(self, lock_name: str) -> None:
        def _d():
            auto_release_time = self.auto_release_info[lock_name]['timeout']
            inspect_interval = auto_release_time / 100
            while True:
                time.sleep(inspect_interval)
                if self.locked(lock_name):
                    if time.time() - self.auto_release_info[lock_name]['last_acquired_time'] >= auto_release_time:
                        self.release_lock(lock_name)

        daemon = threading.Thread(target=_d)
        daemon.daemon = True
        daemon.start()



from tests.__init__ import *
from psplpy.concurrency_utils import ThreadLocks


def tests():
    tl = ThreadLocks()
    lock_name = 'test'
    assert tl.has_lock(lock_name) is False
    assert tl.set_lock(lock_name, auto_release_time=1) is None
    assert tl.has_lock(lock_name) is True
    assert tl.locked(lock_name) is False
    assert tl.acquire_lock(lock_name) is True
    assert tl.locked(lock_name) is True
    assert tl.release_lock(lock_name) is None
    assert tl.locked(lock_name) is False
    print(tl.locks)

    assert tl.acquire_lock(lock_name) is True
    assert tl.acquire_lock(lock_name, timeout=0.1) is False
    print(tl.auto_release_info)
    time.sleep(1)
    assert tl.acquire_lock(lock_name, timeout=1) is True
    print(tl.auto_release_info)


if __name__ == '__main__':
    tests()
